package rabbit.cache;

// $Id: NCache.java,v 1.10 2002/07/22 20:35:09 robo Exp $

import java.io.*;
import java.util.*;
import java.util.zip.*;

import rabbit.util.*;

/** The NCache is like a Hashtable in lookup/insert/delete
 *  The NCache is persistent over sessions (saves itself to disk).
 *  The NCache is selfcleaning, that is it removes old stuff.
 */
public class NCache extends Thread 
{
    private static String DIR = "cache";              // standard dir.
    private static long SIZE  = 10 * 1024 * 1024;     // 10 MB.
    private static long CT = 24 * 60 * 60 * 1000;     // 1 day.

    private static String CACHEINDEX = "cache.index"; // the indexfile.
    private static String TEMPDIR = "temp";
    private static int filesperdir = 256;             // reasonable?

    private boolean changed = false;                  // have we changed?
    private Thread cleaner = null;                    // remover of old stuff.

    private long maxSize = 0;
    private long cacheTime = 0;
    private long fileNo = 0;
    private long currentSize = 0;
    private String dir = null;
    private Hashtable htab = null;                    // mapping Object => CacheEntry.
    private Vector vec = null;

    private File tempdir = null;

    /** Create a cache that uses some dir.
     */
    public NCache () {
	this (DIR);
    }

    /** Create a cache for the given directory.
     * @param dir the name of the directory where the cache should store its files.
     */
    public NCache (String dir) {
	this (dir, SIZE);
    }
    
    /** Create a cache for the given directory with given parameters.
     * @param dir the name of the directory where the cache should store its files.
     * @param maxSize the maximum size of the proxy.
     */
    public NCache (String dir, long maxSize) {
	this (dir, maxSize, CT);
    }
    
    /** Create a cachehandler.
     * @param dir the basedir for the cachehandler.
     * @param maxSize the maximum size of the proxy.
     * @param cacheTime how long the proxy should keep things unless specified otherwise.
     */
    public NCache (String dir, long maxSize, long cacheTime) {
	htab = new Hashtable ();
	vec = new Vector ();
	setMaxSize (maxSize);
	setCacheTime (cacheTime);
	setCacheDir (dir);
	cleaner = new Thread (this, getClass ().getName () + ".cleaner");
	cleaner.setDaemon (true);
	cleaner.start ();
    } 


    /** Get the name of the directory where the cache stores its files.
     * @return the name of the cachedir
     */
    public String getCacheDir () {
	return dir;
    }
    
    /** Sets the cachedir. This will flush the cache and make 
     *  it try to read in the cache from the new dir.
     * @param newDir the name of the new directory to use.
     */
    public synchronized void setCacheDir (String newDir) {
	// save old cachedir.
	if (dir != null)
	    writeCacheIndex ();  
	
	// does new dir exist?
	dir = newDir;
	File dirtest = new File (dir);
	if (!dirtest.exists ()) {
	    dirtest.mkdirs ();
	    if (!dirtest.exists ()) {
		System.err.println ("couldnt create cachedir, exiting");
		System.exit (-1);
	    }
	}
	else if (dirtest.isFile ()) {
	    System.err.println ("Cachedir is a file, that is bad, exiting");
	    System.exit (-1);
	}
	
	tempdir = new File (dir + File.separator + TEMPDIR);
	if (!tempdir.exists ()) {
	    tempdir.mkdir ();
	    if (!tempdir.exists ()) {
		System.err.println ("couldlnt create cache tempdir, exiting");
		System.exit (-1);
	    }
	} 
	else if (tempdir.isFile ()) {
	    System.err.println ("Cache temp dir is a file, that is bad, exiting");
	    System.exit (-1);
	}

	// move to new dir.
	readCacheIndex ();
    }

    /** Get the maximum size for this cache.
     * @return the maximum size in bytes this cache.
     */
    public long getMaxSize () {
	return maxSize;
    }
    
    /** Set the maximum size for this cache.
     * @param newMaxSize the new maximum size for the cache.
     */
    public void setMaxSize (long newMaxSize) {
	maxSize = newMaxSize;
    }

    /** Get the number of miliseconds the cache stores things usually.
     *  This is the standard expiretime for objects, but you can set it for 
     *  CacheEntries individially if you want to.
     *  NOTE 1: dont trust that an object will be in the cache this long.
     *  NOTE 2: dont trust that an object will be removed from the cache when it expires.
     * @return the number of miliseconds objects are stored normally.
     */
    public long getCacheTime () {
	return cacheTime;
    }
    
    /** Set the standard expiry-time for CacheEntries
     * @param newCacheTime the number of miliseconds to keep objects normally.
     */
    public void setCacheTime (long newCacheTime) {
	cacheTime = newCacheTime;
    }
    
    /** Get the current size of the cache
     * @return the current size of the cache in bytes.
     */
    public long getCurrentSize () {
	return currentSize;
    }

    /** Get the current number of entries in the cache.
     * @return the current number of entries in the cache.
     */
    public long getNumberOfEntries () {
	return htab.size ();
    }
    
    /** Get the CacheEntry assosiated with given object.
     * @param o the key.
     * @return the NCacheEntry or null (if not found).
     */ 
    public NCacheEntry getEntry (Object o) {
	NCacheEntry ent =  (NCacheEntry)htab.get (o);
	/*if (ent != null)
	  ent.setVisited (new Date ());*/
	return ent;
    }

    /** Reserve space for a CacheEntry with key o.
     * @param o the key for the NCacheEntry.
     * @return a new CacheEntry initialized for the cache.
     */
    public synchronized NCacheEntry newEntry (Object o) {
	StringBuffer sb = new StringBuffer (50);
	sb.append (dir);
	sb.append (File.separator);
	sb.append (TEMPDIR);
	sb.append (File.separator);
	sb.append (fileNo);
	NCacheEntry entry = new NCacheEntry (o,
					     sb.toString (),
					     fileNo);
	Date exp = new Date (new Date ().getTime () + getCacheTime ()); 
	entry.setExpires (exp);
	fileNo++;
	return entry;
    }

    /** Insert a CacheEntry into the cache.
     * @param ent the CacheEntry to store.
     */
    public void addEntry (NCacheEntry ent) {
	if (ent == null)
	    return;	
	File cfile = new File (ent.getFileName ());
	if (!cfile.exists()) {
	    return;
	}

	long fdir = ent.getId () / filesperdir;
	File f = new File (dir, "" + fdir);
	if (f.exists ()) {
	    if (f.isFile ()) {
		System.err.println ("Doh, wanted cachedir is a file!");
		System.exit (-1);
	    }
	    // good situation...
	} else {
	    f.mkdir ();
	}
	
	StringBuffer sb = new StringBuffer (200);
	sb.append (dir);
	sb.append (File.separator);
	sb.append (fdir);
	sb.append (File.separator);
	sb.append (ent.getId ());
	String newName = sb.toString ();

	File nFile = new File (newName);
	boolean bool = cfile.renameTo (nFile);
	if (bool) {
	    ent.setFileName (newName);	    
	}
	cfile = new File (ent.getFileName ());	
	ent.setSize (cfile.length ());
	ent.setCacheTime (new Date ());

	synchronized (this) {
	    remove (ent.getKey ());
	    htab.put (ent.getKey (), ent);
	    currentSize += ent.getSize ();
	    vec.addElement (ent);
	}
	changed = true;
    }
    
    /** Remove the Entry with key o from the cache.
     * @param o the key for the CacheEntry.
     */
    public void remove (Object o) {
	synchronized (this) {
	    NCacheEntry rent = (NCacheEntry)htab.get (o);
	    if (rent != null) {
		File cfile = new File (rent.getFileName ());
		if (cfile.exists ()) {
		    String parent = cfile.getParent ();
		    cfile.delete ();
		    File p = new File (parent);
		    // humm, until NT does rename in a nice manner check for tempdir.
		    if (p.exists () && !p.equals (tempdir)) {		    
			String ls[] = p.list ();
			if (ls != null && ls.length == 0)
			    p.delete();
		    }
		}
		vec.removeElement (rent);
		currentSize -= rent.getSize ();
	    }
	    htab.remove (o);			
	}
    }

    /** Clear the Cache from files. 
     */
    public void clear () {
	Enumeration e = htab.keys ();
	while (e.hasMoreElements ()) {
	    remove (e.nextElement ());	    
	}
	currentSize = 0;
	changed = true;
    }

    /** Get the CacheEntries in the cache.
     * @return an Enumeration of the CacheEntries.
     */    
    public Enumeration getEntries () {
	return htab.elements ();
    }

    /** Read the info from an old cache.
     */
    private void readCacheIndex () {
	fileNo = 0;
	currentSize = 0;
	htab = new Hashtable ();
	vec = new Vector ();
	try {
	    String name = dir + File.separator + CACHEINDEX;
	    ObjectInputStream is = 
		new ObjectInputStream (new GZIPInputStream (new FileInputStream (name)));
	    fileNo = is.readLong ();
	    currentSize = is.readLong ();
	    htab = (Hashtable)is.readObject ();
	    vec = (Vector)is.readObject ();
	    is.close ();
	} catch (IOException e) {
	    //System.err.println ("Couldnt read " +  dir + File.separator + CACHEINDEX + ", This is bad( but not serius).\nTreating as empty.");
	} catch (ClassNotFoundException e) {
	    //System.err.println ("Couldnt find classes?");
	}
    }
    
    /** Make sure that the cache is written to the disk.
     */
    public void flush () {
	writeCacheIndex ();
    }

    /** Store the cache to disk so we can reuse it later.
     */
    private synchronized void writeCacheIndex () {
	try {
	    String name = dir + File.separator + CACHEINDEX;
	    
	    ObjectOutputStream os = 
		new ObjectOutputStream (new GZIPOutputStream (new FileOutputStream (name)));
	    
	    os.writeLong (fileNo);
	    os.writeLong (currentSize);
	    os.writeObject (htab);
	    os.writeObject (vec);
	    os.close ();
	} catch (IOException e) {
	    System.err.println ("Couldnt write " + dir + File.separator + CACHEINDEX + ", This is serious!\n" + e);
	    System.exit (-1);
	}	
    }

    /** Loop in a cleaning loop.
     */
    public void run () {
	setPriority (Thread. MIN_PRIORITY);
	while (true) {
	    try {
		sleep (15000);		
	    } catch (InterruptedException e) {
		//System.err.println ("Cache interrupted");
	    }
	    
	    // actually for a busy cache this will lag...
	    // but I dont care for now... 
	    Date now = new Date ();
	    long milis = now.getTime ();
	    Enumeration e = htab.elements ();	    
	    while (e.hasMoreElements ()) {
		NCacheEntry ce = (NCacheEntry) e.nextElement ();
		long exp = ce.getExpires ().getTime ();
		if (exp < milis) 
		    remove (ce.getKey ());
	    }

	    // IF SIZE IS TO BIG REMOVE A RANDOM AMOUNT OF OBJECTS.
	    // What we have to be careful about: we must not remove the same
	    // elements two times in a row, this method remove the "oldest" in 
	    // a sense.

	    if (getCurrentSize () > getMaxSize ())
		changed = true;
	    while (getCurrentSize () > getMaxSize ()) 
		remove (((NCacheEntry)vec.firstElement ()).getKey ());
	    
	    if (changed) {
		//System.out.println ("writing " + dir + File.separator + CACHEINDEX + "!");
		writeCacheIndex ();
		changed = false;
	    }
	}
    }

    /** Configure the cache system from the given config.
     * @param config the Property describing the cache settings.
     * @throws IllegalConfigurationException if some setting is strange.
     */
    public void setup (Properties config) throws IllegalConfigurationException {
	if (config == null)
	    return;
	String cachedir = config.getProperty ("directory", "/tmp/rabbit/cache/");
	setCacheDir (cachedir);
	String cmsize = config.getProperty ("maxsize", "10");
	try {
	    setMaxSize (Long.parseLong (cmsize) * 1024 * 1024);     // in MB
	} catch (NumberFormatException e) { 
	    throw (new IllegalConfigurationException ("Bad number for cache maxsize: '" + cmsize + "'"));
	}
	String ctime = config.getProperty ("cachetime", "24");
	try {
	    setCacheTime (Long.parseLong (ctime) * 1000 * 60 * 60); // in hours.
	} catch (NumberFormatException e) { 
	    throw (new IllegalConfigurationException ("Bad number for cache cachetime: '" + ctime + "'"));
	}
    }
}
