package rabbit.filter;

// $Id: DontCacheFilter.java,v 1.5 2001/08/10 21:02:35 robo Exp $

import java.io.*;
import java.net.*;
import java.util.*;

import rabbit.http.*;
import rabbit.proxy.*;
import rabbit.util.*;

/** This is a class that makes sure the proxy doesnt caches certain pages.
 *  It features two types of filtering: either on the URL or the mime type.
 *  
 *  <p><font size=+1><b>URL based matching.</b></font><br>
 *  It uses the config option <tt>dontCacheURLmatching</tt> with a default value
 *  of the empty string. This value is considered a comma (',') separated list.<br>
 *  Matching is done with indexOf so no regular expressions are used for now.<br>
 *  This method should be used on requests, that is append this class to <tt>httpinfilters</tt>
 *  in <tt>rabbit.conf</tt> to use URL matching. 
 *  </p>
 *  <p><font size=+1><b>Mime type based matching.</b></font><br>
 *  It uses the config option <tt>dontCacheMimematching</tt> with a default value
 *  of the empty string. This value is considered a comma (',') separated list.<br>
 *  Matching is done with indexOf so no regular expressions are used for now.<br>
 *  This method should be used on responses, that is append this class to <tt>httpoutfilters</tt>
 *  in <tt>rabbit.conf</tt> to use mime type matching. 
 *  </p> 
 */
public class DontCacheFilter 
{
    private static Properties config = new Properties ();
   
    /** To block creation of this class.
     */
    private DontCacheFilter () {
    }

    /** Test if a socket/header combination is valid or return a new HTTPHeader.
     *  If the request matches a certain criteria dont cache it. This filter is 
     *  good for pages that done send cache-control headers correctly. (chat systems).
     *  This method uses the URL of the request to determine if the resource may be cached.
     * @param socket the Socket that made the request.
     * @param header the actual request made.
     * @param con the Connection handling the request.
     * @return null
     */
    public static HTTPHeader doHTTPInFiltering (Socket socket, HTTPHeader header, Connection con) {
	String dc = config.getProperty ("dontCacheURLmatching");
	String uri = header.getRequestURI ().toLowerCase ();
	if (dc != null) {	    
	    StringTokenizer st = new StringTokenizer (dc, ",");
	    dontCache (con, st, uri);
	}
	
	String oc = config.getProperty ("onlyCacheURLmatching");
	if (oc != null) {
	    StringTokenizer st = new StringTokenizer (oc, ",");
	    onlyCache (con, st, uri);
	}
	return null;
    }

    /** Test if a socket/header combination is valid or return a new HTTPHeader.
     *  If the request matches a certain criteria dont cache it. This filter is 
     *  good for pages that done send cache-control headers correctly. (chat systems).
     *  This method uses the mime type of the response to determine if the resource may be cached.
     * @param socket the Socket that made the request.
     * @param header the actual request made.
     * @param con the Connection handling the request.
     * @return null 
     */
    public static HTTPHeader doHTTPOutFiltering (Socket socket, HTTPHeader header, Connection con) {
	String mimetype = header.getHeader ("content-type").toLowerCase ();
	if (mimetype == null)
	    return null;
	String dc = config.getProperty ("dontCacheMimematching");
	if (dc != null) {
	    StringTokenizer st = new StringTokenizer (dc, ",");
	    dontCache (con, st, mimetype);
	}
	
	String oc = config.getProperty ("onlyCacheMimematching");
	if (oc != null) {
	    StringTokenizer st = new StringTokenizer (oc, ",");
	    onlyCache (con, st, mimetype);
	}
	return null;
    }

    /** if the string toCheck contains any of the tokens in st, do not cache this resource.
     * @param con the Connection handling the request.
     * @param st the tokens we are looking for.
     * @param toCheck the String to check.
     */
    private static void dontCache (Connection con, StringTokenizer st, String toCheck) {
	while (st.hasMoreTokens ()) {
	    if (toCheck.indexOf (st.nextToken ()) > -1) {
		con.setMayCache (false);
		return;
	    }	    
	}
    }
    
    /** if the string toCheck contains any of the tokens in st, cacheing this resource is allowed.
     * @param con the Connection handling the request.
     * @param st the tokens we are looking for.
     * @param toCheck the String to check.
     */
    private static void onlyCache (Connection con, StringTokenizer st, String toCheck) {
	while (st.hasMoreTokens ()) {
	    if (toCheck.indexOf (st.nextToken ()) > -1) {
		return;
	    }	    	    
	}
	con.setMayCache (false);
    }

    /** Setup this class with the given properties.
     * @param properties the new configuration of this class.
     */
    public static void setup (Properties properties) {
	config = properties;
    }
}
