package rabbit.handler;

// $Id: BaseHandler.java,v 1.11 2002/09/14 12:30:03 robo Exp $

import java.io.*;
import java.util.*;
import java.text.*;

import rabbit.cache.*;
import rabbit.http.*;
import rabbit.io.*;
import rabbit.proxy.*;

/** This class is an implementation of the Handler interface.
 *  This handler does no filtering, it only sends the data as
 *  effective as it can.
 */
public class BaseHandler implements Handler
{
    /** The Connection handling the request.*/
    protected Connection con;
    /** The actual request made. */
    protected HTTPHeader request;
    /** The actual response. */
    protected HTTPHeader response;
    /** The stream to read data from. */
    protected InputStream contentstream; 
    /** The stream to send data to. */
    protected MultiOutputStream clientstream;
    
    /** May we cache this request. */
    protected boolean maycache;
    /** May we filter this request */
    protected boolean mayfilter;
    /** The cache entry if available. */
    protected NCacheEntry entry = null;
    /** The cache stream if available. */
    protected OutputStream cacheStream = null;
    /** The length of the data beeing handled or -1 if unknown.*/
    protected long size = -1;

    /** Create a new BaseHansler for the given request.
     * @param con the Connection handling the request.
     * @param request the actual request made.
     * @param response the actual response.
     * @param contentstream the stream to read data from.
     * @param clientstream the stream to write data to.
     * @param maycache May we cache this request? 
     * @param mayfilter May we filter this request?
     * @param size the size of the data beeing handled.
     */
    public BaseHandler (Connection con, 
			HTTPHeader request, 
			HTTPHeader response, 
			InputStream contentstream, 
			MultiOutputStream clientstream, 
			boolean maycache,
			boolean mayfilter, 
			long size) {
	this.con = con;
	this.request = request;
	this.response = response;
	this.contentstream = contentstream;
	this.clientstream = clientstream;
	this.maycache = maycache;
	this.mayfilter = mayfilter;
	this.size = size;
    }
    
    /** Write the response header
     * @throws IOException if writing the response fails.
     */
    protected void writeHeader () throws IOException {
	if (response != null) {
	    clientstream.writeHTTPHeader (response);//response.toString ().getBytes ());
	}
    }

    /** Set up the cache stream if available.
     * @throws IOException if a cachestream couldnt be set up.
     */
    protected void addCacheStream () throws IOException {
	if (maycache) {	    
	    entry = Proxy.getCache ().newEntry (request);
	    String expires = response.getHeader ("Expires");
	    Date exp = null;
	    if (expires != null) {
		exp = HTTPDateParser.getDate (expires); 
		if (exp != null)
		    entry.setExpires (exp);  
		else 
		    Proxy.logError (Proxy.MSG, "unable to parse expire date: '" + expires + 
				    "' for URI: '" + request.getRequestURI () + "'");
	    }
	    entry.setDataHook (response);
	    cacheStream = new FileOutputStream (entry.getFileName ());
	    clientstream.addOutputStream (cacheStream);
	}
    }

    /** This method is used to prepare the stream for the data being sent.
     *  This method does nothing here.
     */
    protected void prepareStream () throws IOException {
	// nothing here.
    }
    
    /** Send the actual data.
     * @throws IOException if reading or writing of the data fails.
     */ 
    protected void send () throws IOException {
	byte v[] = new byte[10240]; // 10kB blocks should be good.
	int read;
	if (size < 0) {
	    while ((read = contentstream.read(v)) > 0) {
		clientstream.write (v, 0, read);
	    }
	} else {
	    long total = 0;
	    while (total < size && (read = contentstream.read(v)) > 0) {
		total += read;
		clientstream.write (v, 0, read);
	    }
	    if (total != size)
		setPartialContent (total, size);
	}
	clientstream.flush ();	
    }

    protected void setPartialContent (long got, long shouldbe) {
	response.setHeader ("RabbIT-Partial", "" + shouldbe);
    }
    
    /** This method is used to finish the stream for the data being sent.
     *  This method does nothing here.
     */
    protected void finishStream () throws IOException {
	// nothing here.
    }

    /** Close nesseccary files and adjust the cached files.
     *  If you override this one, remember to call super.finish ()!
     * @throws IOException if closing the files does.
     */
    protected void finish () throws IOException {
	clientstream.flush ();	
	if (entry != null && maycache) {
	    File f = new File (entry.getFileName ());
	    long filesize = f.length ();
	    entry.setSize (filesize);
	    String cl = response.getHeader ("Conent-Length");
	    if (cl == null) {
		response.removeHeader ("Transfer-Encoding");
		response.setHeader ("Content-Length", "" + filesize);
	    }
	    clientstream.removeOutputStream (cacheStream);
	    cacheStream.close ();
	    Proxy.getCache ().addEntry (entry);
	}
	if (response != null && response.getHeader ("Content-Length") != null)
	    con.setContentLength (response.getHeader ("Content-length"));
	// and close it up....
	con = null;
	request = null;
	response = null;
	contentstream = null;
	clientstream = null;
	entry = null;
	cacheStream = null;
	size = -1;
    }    
    
    /** Handle the request.
     * A request is made in these steps: 
     * <xmp>
     * writeHeader (); 
     * addCacheStream (); 
     * prepareStream ();
     * send (); 
     * finishStream ();
     * finish ();
     * </xmp>
     * @throws IOException if any of the underlying methods does.
     */
    public void handle () throws IOException {
	try {
	    writeHeader ();
	    addCacheStream ();
	    prepareStream ();	
	    send ();
	    finishStream ();
	} catch (IOException e) {
	    removeCache (e);
	    throw e;
	}
	finish ();
    }

    /** Remove the cachestream and the cache entry.
     *  Use this to clean up resources held, but not released due to IOExceptions.
     * @param e the Exception that happened
     */
    protected void removeCache (Exception e) {
	if (cacheStream != null) {
	    File f = new File (entry.getFileName ());
	    if (f.exists ()) {
		Proxy.logError (Proxy.DEBUG, "removing cache file due to exception: " + e);
		f.delete ();
	    }
	    clientstream.removeOutputStream (cacheStream);
	    cacheStream = null;
	    entry = null;
	}
    }
}


