package rabbit.handler;

// $Id: ImageHandler.java,v 1.9 2002/08/21 15:52:28 robo Exp $

import java.io.*;
import java.util.*;

import rabbit.http.*;
import rabbit.io.*;
import rabbit.proxy.*;

/** This class filters images to low quality jpegs.
 */
public class ImageHandler extends BaseHandler
{
    private static final String STD_CONVERT = "/usr/X11R6/bin/convert";
    private static final String STD_CONVERT_ARGS = "-quality 5 $filename[0] jpeg:$filename.c";
    private static Properties config = new Properties ();
    private static boolean doConvert = true;

    private boolean converted = false;
    private long origSize = -1;
    private long lowQualitySize = -1;
    private File convertedFile = null;
    private boolean notEnoughContent = false;
    
    /** Create a new ImageHansler for the given request.
     * @param con the Connection handling the request.
     * @param request the actual request made.
     * @param response the actual response.
     * @param contentstream the stream to read data from.
     * @param clientstream the stream to write data to.
     * @param maycache May we cache this request? 
     * @param mayfilter May we filter this request?
     * @param size the size of the data beeing handled.
     */
    public ImageHandler (Connection con, 
			 HTTPHeader request, 
			 HTTPHeader response, 
			 InputStream contentstream, 
			 MultiOutputStream clientstream,
			 boolean maycache,
			 boolean mayfilter,
			 long size) {
	super (con, request, response, contentstream, clientstream, maycache, mayfilter, size);
	if (size == -1)
	    con.setKeepalive (false);
	con.setChunking (false);
    }
    
    /** Read in the image
     * @throws IOException if reading of the image fails.
     */
    protected void readImage () throws IOException {
	byte[] v = new byte[10240];
	int i;
	long total = 0;
	while ((size < 0 || total < size) && (i = contentstream.read(v)) > 0) {
	    total += i;
	    cacheStream.write (v, 0, i);
	}
	if (size > 0 && size != total) 
	    setPartialContent (total, size);
	cacheStream.close ();
	origSize = new File (entry.getFileName ()).length ();
    }

    /** Convert the image into a low quality jpeg.
     * @throws IOException if conversion fails.
     */
    protected void convertImage () throws IOException {
	String convert = config.getProperty ("convert", STD_CONVERT);
	String convargs = config.getProperty ("convertargs", STD_CONVERT_ARGS);
	
	int idx = 0;
	while ((idx = convargs.indexOf ("$filename")) > -1) {
	    convargs = convargs.substring (0, idx) + entry.getFileName () + convargs.substring (idx + "$filename".length());
	}
	String command = convert + " " + convargs;

	Process ps = Runtime.getRuntime ().exec (command);
	try {
	    ps.waitFor ();
	    if (ps.exitValue () != 0)
		Proxy.logError (Proxy.WARN, "Bad conversion: " + entry.getFileName ());
	} catch (InterruptedException e) {
	    Proxy.logError (Proxy.WARN, "Interupted during wait for: " + entry.getFileName ());
	}

	convertedFile = new File (entry.getFileName () + ".c");
	File iff = new File (entry.getFileName ());
	lowQualitySize = convertedFile.length ();
	if (origSize > lowQualitySize) {
	    response.setHeader ("Content-Type", "image/jpeg");
	    convertedFile.renameTo (new File (entry.getFileName()));
	} else {
	    convertedFile.delete ();
	}
	size = (lowQualitySize < origSize ? lowQualitySize : origSize);
	response.setHeader ("Content-length", "" + size);
	con.setExtraInfo ("imageratio:" + origSize + "/" + lowQualitySize + "=" + ((float)lowQualitySize / origSize));
	
	contentstream.close ();
	contentstream = new FileInputStream (iff);
	convertedFile = null;
    }

    /** Try to convert the image. This is done like this:
     *  <xmp>
     *  super.addCacheStream ();
     *  readImage();
     *  convertImage();
     *  clientstream.removeOutputStream (cacheStream);
     *  </xmp>
     *  We have to use the cachefile to convert the image, and if we
     *  convert it we dont want to write the file to the cache later
     *  on. 
     */
    protected void tryconvert () throws IOException {
	if (doConvert && mayfilter && maycache) {
	    if (size > 2000) {
		super.addCacheStream ();
		readImage ();
		convertImage ();
		converted = true;
		clientstream.removeOutputStream (cacheStream);
	    }	
	} 
    }
    
    /** Write the response header
     * @throws IOException if writing the response fails.
     */
    protected void writeHeader () throws IOException {
	tryconvert ();
	super.writeHeader ();
    }

    /** Set up the cache stream if available.
     * @throws IOException if a cachestream couldnt be set up.
     */
    public void addCacheStream () throws IOException {
	if (!converted)
	    super.addCacheStream ();
    }

    /** clear up the mess we made (remove intermediate files etc).
     * @throws IOException if cleaning fails.
     */
    protected void finish () throws IOException {
	if (convertedFile != null)
	    convertedFile.delete ();
	if (notEnoughContent)
	    removeCache (new Exception ("Content-Length was not met"));
	super.finish ();
    }

    /** Remove the cachestream and the cache entry.
     * @param e the Exception that happened
     */
    protected void removeCache (Exception e) {
	super.removeCache (e);
	if (convertedFile != null)
	    convertedFile.delete ();
    }

    /** Setup this class.
     * @param prop the properties of this class.
     */
    public static void setup (Properties prop) {
	config = prop;
	doConvert = true;
	
	String conv = prop.getProperty ("convert", STD_CONVERT);
	File f = new File (conv);
	if (!f.exists () || !f.isFile()) {
	    Proxy.logError (Proxy.WARN, "convert " + conv + " not found, is your path correct?");
	    doConvert = false;
	}	
    }
}

    
