package rabbit.html;

// $Id: HTMLParser.java,v 1.13 2002/07/08 19:54:06 robo Exp $

import java.util.*;

/** This is a class that is used to parse a block of HTML code into
 *  separate tokens. This parser uses a recursive descent approach.
 */
public class HTMLParser {
    /** The actual data to parse. */
    protected byte[] pagepart;    
    /** The size of the data to parse. */
    protected int length;
    /** The type of the next token. */
    protected int nextToken = START;
    /** Index of the parse. */
    protected int index = 0;
    /** The current tag started here. */
    protected int tagStart = 0;
    /** The current value as a String. */
    protected String stringValue = null;
    /** the current start of string. */
    protected int stringLength = -1;

    /** True if were in a Tag, false otherwise. */
    protected boolean tagmode = false;
    /** The last tag started here. */
    protected int lastTagStart = 0;
    /** The block we have. */
    protected HTMLBlock block;
    
    /** This indicates the start of a block. */
    public final static int START        = 0;
    /** This indicate a String value was found. */
    public final static int STRING       = 1;
    /** This is a Single Quoted String a 'string' */
    public final static int SQSTRING     = 2;
    /** This is a Double Quoted String a &quot;string&quot; */
    public final static int DQSTRING     = 3;
    /** This is the character ''' */
    public final static int SINGELQUOTE  = 4;
    /** This is the character '"' */
    public final static int DOUBLEQUOTE  = 5;
    /** Less Than '<' */
    public final static int LT           = 6;
    /** More Than '>' */
    public final static int MT           = 7;
    /** Equals '=' */
    public final static int EQUALS       = 8;
    /** A HTML comment &quot;&lt;&#33-- some text --&gt;&quot; */
    public final static int COMMENT      = 9;

    /** This indicates the end of a block. */
    public final static int END          = 100;
    /** Unknown token. */
    public final static int UNKNOWN      = 1000;

    /** Create a new HTMLParser */
    public HTMLParser () {
	this.pagepart = null;
    }

    /** Create a new HTMLParser for the given page.
     * @param page the block to parse.
     */
    public HTMLParser (byte[] page) {
	setText (page);
    }

    /** Set the data block to parse.
     * @param page the block to parse.
     */
    public void setText (byte[] page) {
	setText (page, page.length);
    }

    /** Set the data block to parse.
     * @param page the block to parse.
     * @param length the length of the data.
     */
    public void setText (byte[] page, int length) {
	this.pagepart = page;
	this.length = length;
	index = 0;
    }
    
    /** Set the data to parse.
     * @param page the block to parse.
     */ 
    public void setText (String page) {
	setText (page.getBytes ());
    }

    /** Get a String describing the token.
     * @param token the token type (like STRING).
     * @return a String describing the token (like &quot;STRING&quot;)
     */
    protected String getTokenString (int token) {
	switch (token) {
	case START:
	    return "START";
	case STRING:
	    return "STRING";
	case SQSTRING:
	    return "SQSTRING";
	case DQSTRING:
	    return "DQSTRING";
	case SINGELQUOTE:
	    return "SINGELQUOTE";
	case DOUBLEQUOTE:
	    return "DOUBLEQUOTE";
	case LT:
	    return "LT";
	case MT:
	    return "MT";
	case EQUALS:
	    return "EQUALS"; 
	case COMMENT:
	    return "COMMENT";
	case END:
	    return "END";
	case UNKNOWN:
	    return "UNKNOWN";
	default:
	    return "unknown";
	}
    }
    
    /** Scan a String from the block.
     * @throws HTMLParseException if an error occurs.
     * @return STRING
     */
    protected int scanString () throws HTMLParseException {
	int endindex = length;
	int startindex = index - 1;
	
	if (tagmode) {
	loop:
	    while (index < length) {
		// basically a switch ought to be more effective than an if...
		// if there are lots of cases (>4?)
		switch (pagepart[index]) {
		case (int)' ':
		case (int)'\t':
		case (int)'\n':
		case (int)'\r':
		case (int)'\"':
		case (int)'\'':
		case (int)'<':
		case (int)'>':
		case (int)'=':
		    endindex = index;
		    break loop;
		default:
		    index++;
		}
	    }
	} else {
	    while (index < length) {
		if (pagepart[index] == '<') {
		    endindex = index;
		    break;
		} else
		    index++;
	    }
	}
	if (tagmode) {
	    try {
		stringValue = new String (pagepart, startindex, (endindex - startindex), "ISO-8859-1");
	    } catch (Exception e) {
		throw new HTMLParseException ("doh: " + e);
	    }
	} else {
	    stringLength = (endindex - startindex);
	}	    
	return STRING;
    }
    
    /** Scan a quoted tring from the block. The first character is
     *  treated as the quotation character.
     * @throws HTMLParseException if an error occurs.
     * @return SQSTRING, DQSTRING or UNKNOWN (for strange quotes).
     */
    protected int scanQuotedString () throws HTMLParseException {
	int endindex = -1;
	int startindex = index - 1;
	byte start = pagepart[startindex];
	while (index < length) {
	    if (pagepart[index++] == start) {
		endindex = index;
		break;
	    }
	}
	if (endindex == -1) {
	    block.setRest (lastTagStart);
	    return END;
	}
	int l = (endindex < length ? endindex : length) - startindex;
	try {
	    stringValue = new String (pagepart, startindex, l, "ISO-8859-1");
	} catch (Exception e) {
	    throw new HTMLParseException ("doh: " + e);
	}
	switch (start) {
	case (int)'\'':
	    return SQSTRING;
	case (int)'"':
	    return DQSTRING;
	default:
	    return UNKNOWN;
	}
    }

    /** Is this tag a comment?
     * @return true if the block(at current index) starts with !--,
     *  false otherwise.
     */
    protected boolean isComment () {
	int startvalue = index - 1;
	if (index + 3 >= length)
	    return false;
	return (pagepart[index]     == '!' &&
		pagepart[index + 1] == '-' &&
		pagepart[index + 2] == '-');
    }
    
    /** Scan a comment from the block, that is the string up to and
     *	including &quot;-->&quot;.
     * @return COMMENT or END.
     */
    protected int scanComment () throws HTMLParseException {
	int startvalue = index - 1;
	int i = -1;
	int j = index;
	while (j+2 < length) {
	    if (pagepart[j]     == '-' &&
		pagepart[j + 1] == '-' &&
		pagepart[j + 2] == '>') {
		i = j;
		break;
	    }
	    j++;
	}
	if (i > -1) {
	    index = i + 2;
	    nextToken = MT;
	    match (MT);
	    stringLength = index - startvalue;
	    return COMMENT;
	}
	block.setRest (startvalue);
	return END;
    }

    /** Match the token with next token and scan the (new)next token.
     * @param token the token to match.
     * @return the next token.
     */
    protected int match (int token) throws HTMLParseException {
	int ts;
	
	if (nextToken != token)
	    throw new HTMLParseException ("Token: " + getTokenString (token) + " != " + getTokenString (nextToken));
	
	while (index < length) {
	    tagStart = index;
	    stringValue = null;
	    switch (pagepart[index++]) {
	    case (int)' ':
	    case (int)'\t':
	    case (int)'\n':
	    case (int)'\r':
		// TODO evaluate strategy here, a continue here may result 
		// in RabbIT cutting out whitespaces from the html-page.
		continue;
	    case (int)'<':
		ts = tagStart;
		if (isComment ()) {
		    nextToken = scanComment ();
		    tagStart = ts;
		    return nextToken;
		} else 
		    return nextToken = LT;
	    case (int)'>':
		return nextToken = MT;
	    case (int)'=':
		stringValue = "=";
		return nextToken = EQUALS;
	    case (int)'"':
		if (tagmode)
		    return nextToken = scanQuotedString ();	
		// else fallthrough...
	    case (int)'\'':
		if (tagmode)
		    return nextToken = scanQuotedString ();
		// else fallthrough...
	    default:
		return nextToken = scanString ();
	    }
	}

	return nextToken = END;
    }

    /** Scan a value from the block.
     * @return the value or null.
     */
    protected String value () throws HTMLParseException {
	while (nextToken == EQUALS) {
	    match (EQUALS);
	    if (nextToken == STRING ||
		nextToken == SQSTRING || 
		nextToken == DQSTRING) {
		String val = stringValue;
		match (nextToken);
		return val;		   
	    }	    
	    return "";
	}
	return null;
    }

    /** Scan an argument list from the block.
     * @param tag the Tag that have the arguments.
     */
    protected void arglist (Tag tag) throws HTMLParseException {
	String key = null;
	boolean eq = false;
	
	//System.err.println ("parsing arglist for tag: '" + tag + "'");
	while (true) {
	    //System.err.println ("nextToken: " + nextToken + " => " + getTokenString (nextToken));
	    switch (nextToken) {
	    case MT:
		tagmode = false;			
		match (MT);
		return;
	    case STRING:
		key = stringValue;
		match (STRING);
		String value = value ();
		tag.addArg (key, value, false);
		break;
	    case END:
		return;
	    case DQSTRING:
		String ttype = tag.getType ();
		if (ttype != null && ttype.charAt (0) == '!') {
		    // Handle <!DOCTYPE html PUBLIC "-//W3C//DTD HTML 4.01 Transitional//EN">
		    // and similar cases.
		    tag.addArg (stringValue, null, false);
		    match (nextToken);
		} else {
		    //System.err.println ("hmmm, strange arglist..: " + tag);
		    // this is probably due to something like: 
		    // ... framespacing="0"">
		    // we backstep and change the second '"' to a blank and restart from that point...
		    index -= stringValue.length ();
		    pagepart[index] = ' ';
		    match (nextToken);
		    tag.getToken ().setChanged (true);
		}
		break;
	    case LT:   // I consider this an error (in the html of the page) but handle it anyway.
		String type = tag.getLowerCaseType ();	       
		if (type == null ||                             // <<</font.... etc
		    (type.equals ("table") &&    // <table.. width=100% <tr>
		     stringValue == null)) {
		    tagmode = false;
		    return;
		}
		// fall through.
	    default:
		// this is probably due to something like: 
		// <img src=someimagead;ad=40;valu=560>
		// we will break at '=' and split the tag to something like: 
		// <img src=someimagead;ad = 40;valu=560> if we change it.
		// the html is already broken so should we fix it? ignore for now..
		if (stringValue != null)
		    tag.addArg (stringValue, null, false);
		match (nextToken);
	    }
	}
    }

    /** Scan a tag from the block.
     * @param ltagStart the index of the last tag started.
     * @return the Tag scanned.
     */
    protected Tag tag (int ltagStart) throws HTMLParseException {	
	Tag tag = new Tag ();
	Token token = new Token (tag, false);
	
	switch (nextToken) {
	case STRING:
	    tag.setType (stringValue);
	    match (STRING);
	    arglist (tag);
	    if (tagmode) {
		block.setRest (lastTagStart);
	    } else {
		token.setStartIndex (ltagStart);
		block.addToken (token);
	    }
	    break;
	case MT:
	    tagmode = false;			
	    match (MT);
	    break;
	case END:
	    block.setRest (lastTagStart);
	    tagmode = false;	    
	    return null;
	default:
	    arglist (tag);	    
	}
	
	return tag;
    }

    /** Scan a page from the block.
     */
    protected void page () throws HTMLParseException {
	while (true) {
	    switch (nextToken) {
	    case END:
		return;
	    case LT:
		lastTagStart = tagStart;
		tagmode = true;
		match (LT);
		tag (lastTagStart);
		break;
	    case COMMENT: 
		block.addToken (new Token (pagepart, Token.COMMENT, tagStart, stringLength));
		match (COMMENT);
		break;
	    case STRING:
	    default:
		block.addToken (new Token (pagepart, Token.TEXT, tagStart, stringLength));
		match (nextToken);
	    }
	}
    }


    /** Get a HTMLBlock from the pagepart given.
     */
    public HTMLBlock parse () throws HTMLParseException {
	block = new HTMLBlock (pagepart, length);
	nextToken = START;
	match (START);
	page ();
	
	return block;
    }
    
    /** A method to parse a string.
     */
    private static void testParsing (String page) throws Exception {
	//System.out.println (page);
	
	HTMLParser parser = new HTMLParser (page.getBytes());
	HTMLBlock block = parser.parse ();
	String res = "";
	int counter = 0;
	while (block.hasMoreTokens ()) {
	    Token t = block.nextToken ();
	    if (counter++ % 20 == 0) {
		if (t.getType () == Token.TEXT)
		    t.setText ("[" + t.getText () + "]");
		else if (t.getType () == Token.TAG) 
		    t.getTag ().addArg ("is", "new");
	    }
	    //System.out.println ("token is: " + t.getType () + ":" + t.getChanged () + ":" + t.getStartIndex () + ":"+ t);
	    res += t.toString ();
	}
	System.err.println ("original page:");
	System.out.println (page);
	System.err.println ("\n-----------\n");
	System.out.println (res + (block.getRest() != null ? block.getRest () : ""));
	
	System.out.println ("\ntostring on block gives: \n" + block.toString ());
    }

    /** Simple self test function. 
     */
    public static void main (String[] args) {
	try {
	    String page = 
"\n<html>\n<head>\n<title>WebChess</title>\n<body bgcolor=\"#000000\" BACKGROUND=\"\" text=\"#FFFFFF\" LINK=\"#0000FF\" ALINK=\"#000080\" VLINK=\"#008080\">\n<!-- insert loggo here...\n<Img src = \"Images/loga.gif\" alt =\"WebChess\" width = 160 height = 80>\n<B>WebChess</B>\n-->\n<B>WebChess</B> <br>\n<img src = \"Images/guld-genom.gif\" width = 16 height = 16> <a href = \"main.html\" target = \"main\"> About </A> <br>\n\n<img src = \"Images/guld-genom.gif\" width = 16 height = 16> <a href = \"game.html\" target = \"main\"> Using WebChess </A> <br>\n\n<img src = \"Images/guld-genom.gif\" width = 16 height = 16> <a href = \"rules.html\" target = \"main\"> Rules </A><br>\n\n<img src = \"Images/guld-genom.gif\" width = 16 height = 16> <a href = \"history.html\" target = \"main\"> History </A><br>\n\n<img src = \"Images/guld-genom.gif\" width = 16 height = 16> <a href = \"links.html\" target = \"main\"> ChessLinks </A><br>\n\n<img src = \"Images/guld-genom.gif\" width = 16 height = 16> <a href = \"faq.html\" target = \"main\"> Faq </A><br>\n<P>\n<applet code=\"WebChess.Clogin.class\" width = 150 height = 100>\n<param name=Server value=\"web-student.nada.kth.se\">\n<param name=Port value=\"7010\">\n<param name=Imagebase value=\".\">\n</applet>\n\n<hr width = 100%>\n</body>\n<address>\ncreated by <br>\n<a href=\"mailto:d94-rol@nada.kth.se\">Robert Olofsson</a>\n&copy; Copyright Robert Olofsson 1997. All rights reserved.\n</address>\n</html>\nlets test some\ntext\"with quotes\"and so..<hubba>";
	    
	    
	    if (args.length == 0)
		testParsing (page);
	    else 
		for (int i = 0; i < args.length; i++)
		    testParsing (args[i]); 
	    
	} catch (Throwable t) {
	    t.printStackTrace ();
	}
    }
}
