package rabbit.io;

// $Id: ConnectionHandler.java,v 1.9 2002/05/26 18:58:48 robo Exp $

import java.io.*;
import java.net.*;
import java.util.*;

import rabbit.http.*;
import rabbit.proxy.*;
import rabbit.util.*;

/** A class to handle the connections to the net. Should reuse connections if possible.
 */
public class ConnectionHandler extends Thread
{
    // is the cleaner running?
    private boolean running = true;
    
    // The available connections.
    private Hashtable activeConnections;

    // The cleaner thread.
    private Thread cleaner;

    // the keepalivetime.
    private long keepalivetime = 1000;
    
    /** Create a new ConnectionHandler for use. 
     */
    public ConnectionHandler () {
	activeConnections = new Hashtable ();
	Thread cleaner = new Thread (this, getClass().getName () + ".cleaner");
	cleaner.setDaemon (true);
	cleaner.start ();
    }

    /** Check if the cleaner of this ConnectionHandler is running.
     */
    public boolean isCleanerRunning () {
	return running;
    }

    /** Get the connections keept.
     *  @return an Enumeration of Vectors with WebConnections.
     */
    public Enumeration getConnections () {
	return activeConnections.elements ();
    }

    /** Get the addresses we have connections to.
     */
    public Enumeration getAddresses () {
	return activeConnections.keys ();
    }

    /** Get the pool for an Address.
     *  NOTE! synchronize on the pool if you are taking connections from it.
     */
    public Vector getPool (Address a) {
	return (Vector)activeConnections.get (a);
    }
    
    /** A class to handle the addresses of the connections.
     *  Basically just a pair of InetAddress and port number.
     */
    public class Address {
	/** The internet address of this Address. */
	public InetAddress ia;
	/** The port number were connected to. */
	public int port;

	/** Create a new Address with given InetAddress and port 
	 * @param ia the InetAddress this Address is connected to.
	 * @param port the port number this Address is connected to.
	 */
	public Address (InetAddress ia, int port) {
	    this.ia = ia;
	    this.port = port;
	}
	
	/** Get the hash code for this object.
	 * @return the hash code.
	 */
	public int hashCode () {
	    String s = ia.toString () + ":" + port;
	    return s.hashCode ();
	}
	
	/** Compare this objcet agains another object.
	 * @param o the Object to compare against.
	 * @return true if the other Object is an Address connected to 
	 *  the same InetAddress and port, false otherwise.
	 */
	public boolean equals (Object o) {
	    if (o instanceof Address) {
		Address a = (Address)o;
		return (port == a.port && ia.equals (a.ia));
	    }
	    return false;
	}
    }

    /** Get a WebConnection for the given header.
     * @param header the HTTPHeader containing the URL to connect to.
     * @return a WebConnection.
     */
    public WebConnection getConnection (HTTPHeader header) throws IOException {	
	String error = null;
	try {
	    WebConnection wc = null;
	    Proxy.getCounter ().inc ("WebConnections used");
	    String requri = header.getRequestURI ();
	    URL url = new URL (requri);
	    InetAddress ia = Proxy.getInetAddress (url);
	    int port = url.getPort () > 0 ? url.getPort () : 80;
	    port = Proxy.getConnectPort (port);
	    Address a = new Address (ia, port);
	    Vector pool = (Vector)activeConnections.get (a);
	    if (pool == null) {
		return new WebConnection (ia, port);
	    } else {
		synchronized (pool) {
		    if (pool.size () > 0) {
			wc = (WebConnection)pool.elementAt (pool.size () - 1);
			pool.removeElementAt (pool.size () - 1);
		    } else {
			wc = new WebConnection (ia, port);
		    }
		}
		return wc;
	    }	    
	} catch (UnknownHostException e) {
	    Proxy.logError (Proxy.WARN, "Could not find the host: '" + e + "':\n" + header);
	    error = e.toString ();
	} catch (MalformedURLException e) {
	    Proxy.logError (Proxy.WARN, "Bad URI in request: '" + e + "':\n" + header);    
	    error = e.toString ();
	} catch (IllegalArgumentException e) {
	    Proxy.logError (Proxy.WARN, "Bad port number in request: '" + e + "':\n" + header);
	    error = e.toString ();
	}
	throw new IOException ("Could not connect to host: " + error);
    }
    
    /** Return a WebConnection to the pool so that it may be reused.
     * @param wc the WebConnection to return.
     */
    public void releaseConnection (WebConnection wc) {
	if (!wc.getKeepAlive ())
	    return;
	Address a = new Address (wc.getInetAddress (), wc.getPort ());
	wc.setReleased ();
	synchronized (activeConnections) {
	    Vector pool = (Vector)activeConnections.get (a);
	    if (pool == null) {
		pool = new Vector ();
		activeConnections.put (a, pool);
	    }
	    pool.addElement (wc);
	} 
    }
    
    /** Set the keep alive time for this handler.
     * @param milis the keep alive time in miliseconds.
     */
    public void setKeepaliveTime (long milis) {
	keepalivetime = milis;
    }
    
    /** Get the current keep alive time.
     * @return the keep alive time in miliseconds.
     */
    public long getKeepaliveTime () {
	return keepalivetime;
    }
    
    /** The cleaner thread.
     */
    public void run () {
	while (running) {
	    Date d = new Date ();
	    d.setTime (d.getTime () - keepalivetime);
	    Enumeration e = activeConnections.elements ();
	    while (e.hasMoreElements ()) {
		Vector v = (Vector)e.nextElement ();
		synchronized (v) {
		    int vsize = v.size ();
		    for (int i = vsize - 1; i >= 0; i--) {
			WebConnection wc = (WebConnection)v.elementAt (i);
			if (wc.getReleasedAt ().before (d)) {
			    v.removeElementAt (i);
			    wc.close ();
			}
		    }
		}
	    }
	    try {
		sleep (keepalivetime); 
	    } catch (InterruptedException ie) {
		// ignore
	    }
	}       
    }

    /** Simple self test.
     */
    public static void main (String[] args) {	
	try {
	    int times = Integer.parseInt (args[0]);
	    int itimes = Integer.parseInt (args[1]);
	    ConnectionHandler conhandler = new ConnectionHandler ();
	    
	    for (int i = 0; i < times; i++) {
		HTTPInputStream is = new HTTPInputStream (System.in);
		HTTPHeader header = is.readHTTPHeader ();
		
		for (int j = 0; j < itimes; j++) {
		    WebConnection wc = conhandler.getConnection (header);		
		    HTTPOutputStream hos = wc.getOutputStream ();
		    hos.writeHTTPHeader (header);

		    HTTPInputStream his = wc.getInputStream ();		
		    HTTPHeader resp = his.readHTTPHeader ();
		    System.out.write (resp.toString ().getBytes ());
		    int read;
		    if (!header.isHeadOnlyRequest ()) {
			long total = 0;
			byte[] b = new byte[1024];		    
			while (total < wc.dataSize () && (read = his.read (b)) > 0) {
			    total += read;
			    System.out.write (b, 0, read);
			}    
		    }
		    conhandler.releaseConnection (wc);
		}
	    }
	} catch (Exception e) {
	    e.printStackTrace ();
	}
	System.exit (0);
    }

    /** Configure the connection handler system from the given config.
     * @param config the Property describing the cache settings.
     * @throws IllegalConfigurationException if some setting is strange.
     */
    public void setup (Properties config) throws IllegalConfigurationException {
	if (config == null)
	    return;
	String kat = config.getProperty ("keepalivetime", "1000");
	try {
	    setKeepaliveTime (Long.parseLong (kat)); 
	} catch (NumberFormatException e) { 
	    throw (new IllegalConfigurationException ("Bad number for ConnectionHandler keepalivetime: '" + kat + "'"));
	}
    }
}
