package rabbit.io;

// $Id: MultiOutputStream.java,v 1.6 2002/07/08 19:54:12 robo Exp $

import java.io.*;
import java.util.Vector;
import rabbit.http.*;

/** A class to handle automatic writing to several streams simultanius.
 */
public class MultiOutputStream extends OutputStream 
{
    /** how should exceptions be handled?
     *  by closing the stream causing it
     */
    public static final int CLOSE_CURRENT = 1;
    /** Or by respawning the exception up 
     */
    public static final int RESPAWN_EXCEPTION = 2;

    private int mode = CLOSE_CURRENT;
    
    private Vector streams = new Vector ();

    /** Creates a new MultiOutputStream with no 
     *  connected OutputStreams 
     */
    public MultiOutputStream () {	
    }
    
    /** Creates a new MultiOutputStream with one connected OutputStream
     * @param os the OutputStream to connect to
     */
    public MultiOutputStream (OutputStream os) {
	addOutputStream (os);
    }
    
    /** Connects one more OutputStream.
     * @param os the new stream to connect
     */
    public void addOutputStream (OutputStream os) {
	streams.addElement (os);
    }
    
    /** Disconnects one of the underlaying streams.
     * @param os the stream to disconnect.
     */
    public void removeOutputStream (OutputStream os) {
	streams.removeElement (os);
    }

    /** Sets the mode of this stream.
     * @param i one of CLOSE_CURRENT(normal) and RESPAWN_EXCEPTION.
     */
    public void setMode (int i) {
	mode = i;
    }

    /** Write a byte to this stream.
     * @param b the byte to write.
     * @throws IOException if the underlying stream does. 
     */
    public void write(int b) throws IOException {
	int ssize = streams.size ();
	for (int i = 0; i < ssize; i++) {
	    OutputStream os = (OutputStream)streams.elementAt (i);
	    try {
		os.write (b);
	    } catch (IOException e) {
		switch (mode) {
		case RESPAWN_EXCEPTION:
		    throw e;
		default:
		    removeOutputStream (os);
		    ssize--;
		}
	    }
	}	
    }
    
    /** Write a byte array to this stream.
     * @param b the byte array to write.
     * @throws IOException if the underlying stream does. 
     */
    public void write(byte[] b) throws IOException {
	int ssize = streams.size ();
	for (int i = 0; i < ssize; i++) {
	    OutputStream os = (OutputStream)streams.elementAt (i);
	    try {
		os.write (b);
	    } catch (IOException e) {
		switch (mode) {
		case RESPAWN_EXCEPTION:
		    throw e;
		default:
		    removeOutputStream (os);
		    ssize--;
		}
	    }
	}	
    }
    
    /** Write a byte array to this stream.
     * @param b the byte to write.
     * @param off the starting offset.
     * @param len the number of bytes to write.
     * @throws IOException if the underlying stream does. 
     */
    public void write(byte[] b, int off, int len) throws IOException {
	int ssize = streams.size ();
	for (int i = 0; i < ssize; i++) {
	    OutputStream os = (OutputStream)streams.elementAt (i);
	    try {
		os.write (b, off, len);
	    } catch (IOException e) {
		switch (mode) {
		case RESPAWN_EXCEPTION:
		    throw e;
		default:
		    removeOutputStream (os);
		    ssize--;
		}
	    }
	}	
    }

    /** Write a HTTPHeader on this stream.
     * @param header the HTTPHeader to write.
     * @throws IOException if the header could not be written correctly.
     */
    public void writeHTTPHeader (HTTPHeader header) throws IOException {
	byte[] b = null; 
	int ssize = streams.size ();
	for (int i = 0; i < ssize; i++) {
	    OutputStream os = (OutputStream)streams.elementAt (i);
	    try {		    
		if (os instanceof HTTPOutputStream) {
		    ((HTTPOutputStream)os).writeHTTPHeader (header);
		} else {
		    if (b == null)
			b = header.toString ().getBytes ();
		    os.write (b);
		}
	    } catch (IOException e) {
		switch (mode) {
		case RESPAWN_EXCEPTION:
		    throw e;
		default:
		    removeOutputStream (os);
		    ssize--;
		}
	    }
	}
    }
    
    /** Write any pending data.
     * @throws IOException if the underlying stream does. 
     */
    public void flush() throws IOException {
	int ssize = streams.size ();
	for (int i = 0; i < ssize; i++) {
	    OutputStream os = (OutputStream)streams.elementAt (i);
	    try {
		os.flush ();
	    } catch (IOException e) {
		switch (mode) {
		case RESPAWN_EXCEPTION:
		    throw e;		    
		default:
		    removeOutputStream (os);
		    ssize--;
		}
	    }
	}    
    }
    
    /** Close this stream.
     * @throws IOException if the underlying stream does. 
     */
    public void close() throws IOException {
	int ssize = streams.size ();
	for (int i = 0; i < ssize; i++) {
	    OutputStream os = (OutputStream)streams.elementAt (i);
	    try {
		os.close ();
	    } catch (IOException e) {
		switch (mode) {
		case RESPAWN_EXCEPTION:
		    throw e;
		default:
		    removeOutputStream (os);
		    ssize--;
		}
	    }
	}	
    }

    /** Simple self test
     */
    public static void main (String args[]) {
	try {
	    MultiOutputStream multi = new MultiOutputStream ();
	    
	    FileOutputStream fis1 = new FileOutputStream ("hepp");
	    FileOutputStream fis2 = new FileOutputStream ("hopp");
	    FileOutputStream fis3 = new FileOutputStream ("happ");
	    multi.addOutputStream (fis1);
	    multi.addOutputStream (fis2);
	    multi.addOutputStream (fis3);
	    
	    PrintWriter p = new PrintWriter (multi);
	    p.println("tjo och hej i lingonskogen");
	    p.flush ();
	    p.close ();
	} catch (Exception e) {
	    System.out.println ("An error occured: " + e);
	} 
    }
}
