package rabbit.io;

// $Id: WebConnection.java,v 1.7 2001/09/02 13:37:22 robo Exp $

import java.io.*;
import java.net.*;
import java.util.*;

import rabbit.http.*;
import rabbit.proxy.*;
import rabbit.util.*;

/** A class to handle a connection to the Internet.
 */
public class WebConnection 
{
    private Socket socket;
    private HTTPInputStream webis;
    private HTTPOutputStream webos;
    private InetAddress ia;
    private int port;
    private Date releasedAt;

    /** Create a new WebConnection to the given InetAddress and port.
     * @param InetAddress the computer to connect to.
     * @param port the port number to connect to.
     */
    public WebConnection (InetAddress ia, int port) throws IOException {
	this.ia = ia; 
	this.port = port;
	try {
	    socket = new Socket (ia, port);	
	    socket.setSoTimeout (5*60*1000);        // 5 min should be enought.
	    Proxy.getCounter ().inc ("WebConnections created");
	    webos = new HTTPOutputStream (socket.getOutputStream ());
	    webis = new HTTPInputStream (socket.getInputStream ()); 
	} catch (IOException e) {
	    Proxy.logError (Proxy.WARN, "couldnt connect to the remote site (" + e.getMessage () + ")");
	    throw new IOException ("Could not connect to the site: " + ia + ":" + port);
	}
    }

    /** Close the connection.
     */
    public void close () {
	try {
	    webos.close ();
	    webos = null;
	    webis.close ();
	    webis = null;
	    socket.close ();
	    socket = null;
	    ia = null;
	    releasedAt = null;
	} catch (IOException e) {
	    Proxy.logError (Proxy.WARN, "couldnt close WebConnection: " + e.getMessage ());
	}
    }
    
    /** Get the InetAddress that this WebConnection is connected to.
     * @return the InetAddress.
     */
    public InetAddress getInetAddress () {
	return ia;
    }
    
    /** Get the port number this WebConnection is connected to.
     * @return the port number.
     */
    public int getPort () {
	return port;
    }

    /** Mark this WebConnection as released at current time.
     */
    public void setReleased () {
	setReleased (new Date ());
    }

    /** Mark this WebConnection as released at given time.
     * @param d the time that this WebConnection is released.
     */
    public void setReleased (Date d) {
	try {
	    webis.finish ();
	    webos.finish ();
	} catch (IOException e) {
	    Proxy.logError (Proxy.WARN, "could not finish streams: " + e);	    
	}
	releasedAt = d;
    }
    
    /** Get the time that this WebConnection was released.
     */
    public Date getReleasedAt () {
	return releasedAt;
    }

    /** Get the InputStream.
     * @return an HTTPInputStream.
     */
    public HTTPInputStream getInputStream () {
	return webis;
    }
    
    /** Get the OutputStream of this WebConnection.
     * @return an HTTPOutputStream.
     */
    public HTTPOutputStream getOutputStream () {
	return webos;
    }

    /** Get the keepalive value of this WebConnection.
     * @return true if this WebConnection may be reused.
     */
    public boolean getKeepAlive () {
	return webis.getKeepAlive ();
    }

    /** Get the size of the data being fetched.
     * @return the number of bytes of the page or -1 if unknown.
     */
    public long dataSize () {
	return webis.dataSize ();
    }

    /** Is this request chunked?
     * @return true if the last read request was chunked, false otherwise.
     */
    public boolean chunked () {
	return webis.chunked ();
    }

    /** Get the last chunked page.
     * @return an InputStream to the last chunked page.
     */
    public InputStream getChunkStream () {
	return webis.getChunkStream ();
    }

    /** Write an HTTP header on this connection. 
     *  This is a convenience function to do <xmp>
     *  HTTPOutputStream hos = getOutputStream ();
     *  hos.writeHTTPHeader (header);
     *  </xmp>
     * @param header the Header to write.
     * @throw IOException if an I/O error occurs.
     */
    public void writeHTTPHeader (HTTPHeader header) throws IOException {
	webos.writeHTTPHeader (header);
    }
}
