package rabbit.linkchecker;

// $Id: LinkVerifier.java,v 1.9 2001/11/11 20:02:51 robo Exp $

import java.net.*;
import java.util.*;

import rabbit.io.*;
import rabbit.html.*;
import rabbit.http.*;
import rabbit.util.*;


/** linkchecker...
 *  This is how it should work when it is finished:
 * <pre>
 * mailto:user@host.net => SMTP, VRFY user
 * href=page => follow the link if page in given set (test that external link works but do not recurse them).
 * news => ignore. 
 * file => warn
 * ftp  => check that the file exist.
 * 
 * imagemaps => as specified above.
 * frameset  => as specified above.
 * applets   => check that it exists.
 * images    => check that it exists.
 * anchors   => check that the page has them.
 *
 * proxy support.
 * </pre>
 * 
 * authentifiering => if basic-auth then use it since it is simple.
 */
public class LinkVerifier extends Thread
{
    /** Current version */
    public static final String VERSION = "RabbIT linkverifier version 1.0a";

    /** The web connection handler. */
    private static ConnectionHandler conhandler = new ConnectionHandler ();

    /** The urls we have fetched. */
    private static Hashtable history = new Hashtable ();
    
    /** The config we are using. */
    private static Config cfg;

    /** The hosts to recurse */
    private static Hashtable recursehosts = new Hashtable ();

    /** The vector of loggers. */
    private static Vector loggers = new Vector ();

    /** The url this instance should get. */
    private URL urltoget;
    
    /** Add an url to the list of already fetched urls.
     * @param old the URL we come from
     * @param url the URL to get.
     * @return true if it is ok to get the new url, false otherwise.
     */
    private static boolean addToGet (URL old, URL url) {
	synchronized (history) {
	    //System.err.println ("checking: " + old + " => " + url);
	    Object gotten = history.get (url);
	    if (gotten != null) {
		//System.err.println ("already gotten:" + old);
		return false;
	    }

	    // insert it into history is not bad even if it is never fetched...
	    history.put (url, url);
	    //System.err.println ("checking: '" + old.getHost () + 
	    // "' => '" + url.getHost () + "'");
	    Object rechost = recursehosts.get (old.getHost ());
	    if (rechost == null)
		return false;
	    //System.err.println ("seems ok to get:" + url + " <= (" + old + ")");
	    return true;
	}
    }
    
    /** Create a new LinkVerifier for the given urls.
     * @param prev the URL we came from.
     * @param url the URL to fetch.
     */
    public LinkVerifier (URL prev, URL url) {
	urltoget = url;
	start ();
    }    

    /** Fetch the page and recurse the urls found.
     */
    public void run () {
	try {
	    sleep (1000);
	    HTTPHeader header = new HTTPHeader ();
	    header.setRequestLine ("GET " + urltoget + " HTTP/1.1");
	    header.setHeader ("Host", urltoget.getHost ());
	    WebConnection wc = conhandler.getConnection (header);		
	    HTTPOutputStream hos = wc.getOutputStream ();
	    
	    hos.writeHTTPHeader (header);
	    
	    HTTPInputStream his = wc.getInputStream ();		
	    HTTPHeader resp = his.readHTTPHeader ();
	    
	    if (!(resp.getStatusCode ().indexOf ("200") > -1)) {
		log (urltoget, resp.getStatusLine ());
		return;
	    }
	    String contype = resp.getHeader ("Content-Type");
	    if (contype == null || !contype.equals ("text/html")) {
		log (urltoget, resp);
		return;
	    }
	    int read;
	    HTMLParser parser = new HTMLParser ();
	    HTMLBlock block = null;
	    
	    long total = 0;
	    byte[] v = new byte[1024];
	    
	    if (wc.chunked ()) {
		his = new HTTPInputStream (wc.getChunkStream ());
	    }
	    
	    while (total < wc.dataSize () && (read = his.read (v)) > 0) {
		total += read;
		byte[] b = new byte [(block != null ? block.restSize () : 0) + read];
		if (block != null) 
		    block.insertRest (b);
		System.arraycopy (v, 0, b, (block != null ? block.restSize () : 0), read);	    
		parser.setText (b);
		try {
		    block = parser.parse ();
		    Vector tokens = block.getTokens ();
		    for (int a = 0; a < tokens.size (); a++) {
			Token t = (Token)tokens.elementAt(a);
			if (t.getType () == Token.TAG) {
			    Tag tag = t.getTag ();
			    String nurl = null;
			    if (tag.getLowerCaseType ().equals ("a") || 
				tag.getLowerCaseType ().equals ("area")) {
				nurl = tag.getAttribute ("href");
			    }
			    else if (tag.getLowerCaseType ().equals ("frame") || 
				     tag.getLowerCaseType ().equals ("img")) {
				nurl = tag.getAttribute ("src");
			    }
			    if (nurl != null) 
				handleURL (urltoget, nurl);
			}
		    }
		} catch (HTMLParseException e) {
		    System.err.println ("bad HTML: " + e.toString());
		}
	    }
	    log (urltoget, resp);
	    conhandler.releaseConnection (wc);
	} catch (Exception e) {
	    log (urltoget, e.toString ());
	    return;
	}
    }
    
    private static void handleURL (URL urltoget, String nurl) {
	if (nurl.startsWith ("\"") || nurl.startsWith ("'"))
	    nurl = nurl.substring (1);
	if (nurl.endsWith ("\"") || nurl.endsWith ("'"))
	    nurl = nurl.substring (0, nurl.length () - 1);
	// cut away anchors....
	int i = nurl.indexOf ("#");
	if (i > -1)
	    nurl = nurl.substring (0, i);
	URL turl;
	try {
	    turl = new URL (urltoget, nurl);
	    if (addToGet (urltoget, turl)) {
		String proto = turl.getProtocol ();
		if (proto != null) {		
		    proto = proto.toLowerCase ();
		    if (proto.equals ("http"))
			new LinkVerifier (urltoget, turl);
		    else if (proto.equals ("ftp")) {
			System.err.println ("not able to check ftp sites yet: " + 
					    turl + " from " + urltoget);
		    }
		    else if (proto.equals ("mailto")) {
			System.err.println ("not able to check mailto yet: " + 
					    turl + " from " + urltoget);
		    }
		    else if (proto.equals ("news")) {
			System.err.println ("ignoring news sites: " + 
					    turl + " from " + urltoget);
		    } 
		    else if (proto.equals ("file")) {
			System.err.println ("file:// is a bad idea...: " + 
					    turl + " from " + urltoget);
		    }
		    else {
			System.err.println ("protocoll: " + 
					    proto + 
					    " is unknown to me. Ignoring");
		    }
		}
	    }
	} catch (MalformedURLException e) {
	    System.err.println ("bad url:" + e);
	    // ignore...
	}	
    }
    
    private static void setup () {
	StringTokenizer st = 
	    new StringTokenizer (cfg.getProperty ("", "recursehosts", ""), ",");
	while (st.hasMoreElements ()) {
	    String host = st.nextToken ().trim ();
	    recursehosts.put (host, host);
	}
	st = new StringTokenizer (cfg.getProperty ("", 
						   "loggers", 
						   "rabbit.linkverifier.SimpleLogger"), ",");
	while (st.hasMoreElements ()) {
	    String cls = st.nextToken ().trim ();
	    try {
		Class rcls = Class.forName (cls);
		Object logger = rcls.newInstance ();
		System.err.println ("adding " + cls + " as logger");
		loggers.addElement (logger);
	    } catch (Exception e) {
		System.err.println ("Class: " + cls + 
				    " could not be used as a logger: \n" + e);
	    }
	}
    }

    private static synchronized void log (URL url, String msg) {
	for (int i = 0; i < loggers.size (); i++) {
	    Logger l = (Logger)loggers.elementAt (i);
	    l.log (url, msg);
	}
    }

    private static synchronized void log (URL url, HTTPHeader header) {
	for (int i = 0; i < loggers.size (); i++) {
	    Logger l = (Logger)loggers.elementAt (i);
	    l.log (url, header);
	}
    }

    private static synchronized void startLogging () {
	for (int i = 0; i < loggers.size (); i++) {
	    Logger l = (Logger)loggers.elementAt (i);
	    l.startLogging (cfg);
	}
    }

    private static synchronized void endLogging () {
	for (int i = 0; i < loggers.size (); i++) {
	    Logger l = (Logger)loggers.elementAt (i);
	    l.endLogging ();
	}
    }

    /** print out the helptext to the user.
     */
    public static void printHelp () {
	System.out.println ("java rabbit.linkchecker.LinkVerifier [(-f|--file) <configfile>] [url]*");
	System.out.println ("will use conf/linkchecker.conf if no configfile is specified");
    }


    public static void main (String[] args) {
	try {
	    String configfile = "conf/linkchecker.conf";
	    int i = 0;
	    Vector urls = new Vector ();
	    while (args.length > i) {		
		// use custom ini-file?
		if (args[i].equals ("-f") || 
		    args[i].equals ("--file")) {
		    i++;
		    if (args.length > i) {
			configfile = args[i];
		    }
		    else {
			System.err.println ("No config file specified, exiting.");
			System.exit (-1);
		    }
		}
		
		// user wants help??
		else if (args[i].equals ("-?") || 
			 args[i].equals("-h") || 
			 args[i].equals ("--help")) {
		    printHelp();
		    System.exit (0);
		} 

		// version ?
		else if (args[i].equals ("-v") || args[i].equals ("--version")) {
		    System.out.println (VERSION);
		    System.exit (0);
		}	
		
		// else? stack it up.
		else {
		    urls.addElement (args[i]);
		}
		i++;
	    }
	    
	    cfg = new Config (configfile);
	    setup ();
	    if (urls.size () == 0)
		printHelp ();
	    startLogging ();
	    for (i = 0; i < urls.size (); i++)
		//new LinkVerifier (null, new URL ((String)urls.elementAt (i)));
		handleURL (new URL ((String)urls.elementAt (i)), "");
	} catch (Exception e) {
	    e.printStackTrace ();
	}
    }
}

