package rabbit.proxy;

// $Id: Connection.java,v 1.37 2002/09/15 12:04:34 robo Exp $

import java.io.*;
import java.lang.reflect.*;
import java.net.*;
import java.util.*;

import rabbit.cache.*;
import rabbit.handler.*;
import rabbit.http.*;
import rabbit.io.*;
import rabbit.meta.*;
import rabbit.util.*;


/** This is the class that handles one connection 
 *  And make sure the content is delivered to the client.
 *  It reads the request and get the data (from the cache or the web), 
 *  and filters it and sends the data to the client. If Keepalive is suitable
 *  it reads the next request.
 */
public class Connection extends RestartableThread 
{
    /** The current status of this connection. */
    private String   status;
    private Date     started = new Date ();

    private Socket             socket = null;
    private HTTPInputStream    in     = null;
    private HTTPOutputStream   client = null;
    private MultiOutputStream  out    = null;
    
    private boolean  keepalive = true;
    private boolean  meta      = false;
    private int      requests  = 0;
    
    private boolean  chunk       = true;
    private boolean  mayusecache = true;
    private boolean  maycache    = true;
    private boolean  mayfilter   = true;
    
    /** If the user has authenticated himself */
    private String username = null;
    private String password = null;

    private String requestline   = null;
    private String statuscode    = null;
    private String extrainfo     = null;
    private String contentlength = null;

    private boolean chunkedInput = false;

    private static long counter = 0;
    private static Hashtable metaHandlers = new Hashtable ();

    /** Create a new Connection.
     *  You need to call setsocket on this connection to make it do stuff.
     */
    public Connection () {
	super ("RabbIT2: " + counter++);
    }

    /** Create a new Connection for a given socket.
     * @param socket the Socket that is requesting service.
     */
    public Connection (Socket socket) {
	super ("RabbIT Connection: " + counter++);
	setSocket (socket);
    }
    
    /** Give this connection a socket to handle
     * @param socket the Socket that is requesting service.
     */
    public void setSocket (Socket socket) {
	this.socket = socket;
	keepalive = true;
	meta      = false;
	requests  = 0;

	if (!checkIPAccess (socket)) {
	    Proxy.logError (Proxy.WARN, "Rejecting access from " + socket.getInetAddress ());
	    Proxy.getCounter ().inc ("Rejected IP:s");
	    closeDown ();
	}
	else {
	    start ();
	}
    }

    /** Close down nicely.
     */
    protected void closeDown () {
	try {
	    in.close ();
	    out.close ();
	    socket.close ();
	} catch (IOException e) {
	    Proxy.logError (Proxy.WARN, "Problems to close the sockets are usually a bad thing..." + e);
	}
	in = null;
	out = null;
	socket = null; 
	status = "idle";
	Proxy.removeConnection (this);
    }

    /** Check if this socket is in the possible range of ip:s being served.
     * @param socket the Socket to check.
     * @return true if the socket is allowed, false if it should be rejected.
     */
    public boolean checkIPAccess (Socket socket) {
	Vector v = Proxy.accessfilters;
	Object[] params = new Object[1];
	params[0] = socket;
	try {
	    int vsize = v.size ();
	    for (int i = 0; i < vsize; i++) {
		Method m = (Method)v.elementAt (i);
		if (((Boolean)m.invoke (m.getDeclaringClass(), params)).booleanValue())
		    return true;
	    }
	} catch (InvocationTargetException e) {
	    Proxy.logError (Proxy.WARN, "checkIPAccess: invocation failed" + e);
	} catch (IllegalAccessException e) {
	    Proxy.logError (Proxy.WARN, "checkIPAccess: invocation failed" + e);	    
	}
	return false;
    }

    /** Check if this SSL request is allowed and handle it
     *  accordingly.
     * @param header the SSL request.
     */
    private void checkAndHandleSSL (HTTPHeader header) {
	if (sslIsAllowed (header)) {
	    handleSSL (header);
	} else {
	    HTTPHeader badresponse = get403 ();
	    send (badresponse);
	    setKeepalive (false);
	    statuscode = badresponse.getStatusCode ();
	}
	status = "logging connection";
	Proxy.logConnection (this);
    }

    /** Resets the statuses for this connection.
     */
    private void clearStatuses () {
	status      = "Reading request";
	started     = new Date ();
	username    = null;
	password    = null;
	chunk       = true;
	mayusecache = true;
	maycache    = true;
	mayfilter   = true;
	requestline = "?";
	statuscode  = "200";
	extrainfo   = null;
	contentlength = "-";
	chunkedInput = false;
    }
    
    /** Handles the case where the request does not have a valid body.
     * @param header the request made.
     */
    private void handleBadContent (HTTPHeader header) {
	Proxy.logError (Proxy.DEBUG, "bad content for:\n" + header);
	doError (400, "Bad content");
	statuscode = "400";
	status = "logging connection";
	Proxy.logConnection (this);
    }

    /** Filter the request and handle it. 
     * @param header the request
     */
    private void filterAndHandleRequest (HTTPHeader header) {
	// filter the request based on the header.
	// a bad response means that the request is blocked (could be ad blocking)...
	HTTPHeader badresponse = filterHTTP (Proxy.httpinfilters, header);	
	if (badresponse != null) {
	    send (badresponse);
	    setKeepalive (false);
	    statuscode = badresponse.getStatusCode ();
	} else {
	    status = "Handling request";
	    if (getMeta ()) {
		// Ok, so theese pages we dont know the size of...
		setKeepalive (false);
		handleMeta (header);
	    } else {
		handleRequest (header);
	    }
	}
    }

    private void logExceptionAndSendError (Throwable t) {
	// Log the Exception....
	StringWriter sw = new StringWriter ();
	PrintWriter sos = new PrintWriter (sw);
	t.printStackTrace (sos);
	Proxy.logError (Proxy.ERROR, sw.toString ());
	// Tell the client what happened.
	HTTPHeader err = get500 (t);
	send (err);
    }   
    
    /** Handle the incomming requests for as long as we can keep the connection alive.
     */
    public void doWork () {
	try {
	    do {
		clearStatuses ();
		requests++;
		if (!setupStreams ())
		    break;
		
		// first read the header....
		HTTPHeader header = null;
		try {
		    header = in.readHTTPHeader ();
		    String requestVersion = header.getHTTPVersion ().toUpperCase ();
		    header.addHeader ("Via", requestVersion + " RabbIT");
		} catch (InterruptedIOException e) {
		    Proxy.logError (Proxy.INFO, "Keepalive timed out: " + e);
		    break;
		} catch (IOException e) {
		    Proxy.logError (Proxy.INFO, 
				    "" + socket.getInetAddress ().getHostAddress () + 
				    " disconnected when reading header\n" + e);
		    break;
		}

		requestline = header.getRequestLine ();
		Proxy.getCounter ().inc ("Requests");		

		// SSL requests are special in a way... 
		// Dont depend upon being able to build URLs from the header...
		if (isSSLRequest (header)) {
		    checkAndHandleSSL (header);
		    break;
		}

		// Read in any posted data.
		if (header.getHeader ("Content-length") != null) {
		    setMayUseCache (false);
		    setMayCache (false);
		    if (!readContent (header)) {
			handleBadContent (header);
			break;
		    }
		}
		
		String te = header.getHeader ("Transfer-Encoding");
		if (te != null && te.equalsIgnoreCase ("chunked")) {
		    setMayUseCache (false);
		    setMayCache (false);
		    header.setContentStream (in);
		}

		filterAndHandleRequest (header);
		status = "logging connection";
		Proxy.logConnection (this);
	    } while (keepalive);
	} catch (Throwable t) {
	    logExceptionAndSendError (t);
	}
	closeDown ();
    }

    /** Set up the streams used for reading request and sending data to and from the client.
     * @return true if all goes well.
     */
    protected boolean setupStreams () {
	// the streams are initated to null and 
	// upon keepalive they are already setup.
	if (in != null && out != null)
	    return true;
	try {
	    in = new HTTPInputStream (new BufferedInputStream (socket.getInputStream ()));
	    client = new HTTPOutputStream (socket.getOutputStream ());
	    out = new MultiOutputStream (client);
	} catch (IOException e) {
	    try {
		socket.close ();
		socket = null;
	    } catch (IOException e2) {
		Proxy.logError (Proxy.FATAL, "Not being able to close the sockets will lead to system shutdown.");
		System.exit (-4711);
	    }
	    Proxy.logError (Proxy.ERROR, "Couldnt get streams from socket: " + e);
	    return false;
	}
	return true;
    }

    /** If this request has a body (due to some post or so)
     *  read it in and append it to the request.
     */
    protected boolean readContent (HTTPHeader header) {
	String sl = (String)header.getHeader("Content-length");
	int contentLength = Integer.parseInt (sl.trim ());

	byte v[] = new byte[contentLength];
	try {
	    in.readFully (v, 0, contentLength);   // well we should read this many!
	    header.setContent (new String (v));
	} catch (EOFException e) {
	    Proxy.logError (Proxy.WARN, "Couldnt read all content");
	    return false;
	} catch (IOException e) {
	    Proxy.logError (Proxy.INFO, "Unexpected termination:" + e);
	    return false;
	}
	return true;
    }

    /** Filter the headers using the methods in the vector.
     * @param filters a Vector with Methods.
     * @param in the request or response header.
     * @return null if all is ok, a HTTPHeader if this request is blocked.
     */
    public HTTPHeader filterHTTP (Vector filters, HTTPHeader in) {
	Object params[] = new Object[3];
	params[0] = socket;
	params[1] = in;
	params[2] = this;
	
	try {
	    int fsize = filters.size ();
	    for (int i = 0; i < fsize; i++) {
		Method m = (Method)filters.elementAt (i);
		HTTPHeader badresponse = (HTTPHeader)m.invoke (m.getDeclaringClass(), params);
		if (badresponse != null)
		    return badresponse;
	    }
	} catch (InvocationTargetException e) {
	    Proxy.logError (Proxy.WARN, "filterHTTP: invocation failed: " + e);
	    e.printStackTrace ();
	} catch (IllegalAccessException e) {
	    Proxy.logError (Proxy.WARN, "filterHTTP: invocation failed: " + e);	    
	    e.printStackTrace ();
	}
	return null;
    }

    /** Send a header to the client.
     * @param header the HTTPHeader to send.
     */
    protected void send (HTTPHeader header) {
	try {
	    out.writeHTTPHeader (header);
	    out.flush ();
	} catch (IOException e) {
	    setKeepalive (false);
	}
    }
    
    /** Get a new Handler for the given request made.
     * @param cls the Class of the handler to create.
     * @param header the request.
     * @param webheader the response.
     * @param contentStream the data to read.
     * @param out the client (and possibly the cache).
     * @param maycache if the handler may cache the response.
     * @param mayfilter if the handler may filter the response.
     * @param size the Size of the data beeing handled (-1 = unknown length).
     */
    protected Handler getNewInstance (Class cls,
				      HTTPHeader header, 
				      HTTPHeader webheader, 
				      InputStream contentStream, 
				      MultiOutputStream out, 
				      boolean maycache, 
				      boolean mayfilter, 
				      long size) {
	try {
	    Class[] paramtypes = new Class[8];
	    paramtypes[0] = getClass ();
	    paramtypes[1] = HTTPHeader.class;
	    paramtypes[2] = HTTPHeader.class;
	    paramtypes[3] = InputStream.class;
	    paramtypes[4] = MultiOutputStream.class;
	    paramtypes[5] = boolean.class;
	    paramtypes[6] = boolean.class;
	    paramtypes[7] = long.class;

	    Object[] args = new Object[8];
	    args[0] = this;
	    args[1] = header;
	    args[2] = webheader;
	    args[3] = contentStream;
	    args[4] = out;
	    args[5] = new Boolean (maycache);
	    args[6] = new Boolean (mayfilter);
	    args[7] = new Long (size);
	    
	    Constructor c = cls.getConstructor(paramtypes);
	    return (Handler)c.newInstance (args);
	} catch(NoSuchMethodException e) {
	    return null;
	} catch (SecurityException e) {
	    return null;
	} catch (InstantiationException e) {
	    return null;
	} catch (IllegalAccessException e) {
	    return null;
	} catch (InvocationTargetException e) {
	    return null;
	}
    }

    private boolean checkStrongEtag (String et, String im) {
	return !im.startsWith ("W/") && im.equals (et);
    }

    private boolean checkWeakEtag (HTTPHeader h1, HTTPHeader h2) {
	String et1 = h1.getHeader ("Etag");
	String et2 = h2.getHeader ("Etag");
	if (et1 == null || et2 == null)
	    return true;
	return checkWeakEtag (et1, et2);
    }

    private boolean checkWeakEtag (String et, String im) {
	if (et.startsWith ("W/"))
	    et = et.substring (2);
	if (im.startsWith ("W/"))
	    im = im.substring (2);
	return im.equals (et);
    }

    public HTTPHeader checkIfMatch (HTTPHeader header, NCacheEntry entry) {
	if (entry == null)
	    return null;
	String im = header.getHeader ("If-Match");
	if (im == null)
	    return null;
	HTTPHeader oldresp = (HTTPHeader)entry.getDataHook ();
	HTTPHeader expfail = checkExpectations (header, oldresp);
	if (expfail != null)
	    return expfail;
	String et = oldresp.getHeader ("Etag");
	if (!checkStrongEtag (et, im))
	    return get412 ();
	return null;
    }

    /** Check if the request allows us to use a "304 Not modified" response.
     * @param in the request being made.
     * @param entry the cache entry.
     */
    public HTTPHeader is304 (HTTPHeader in, NCacheEntry entry) {
	if (entry == null)
	    return null;
	HTTPHeader oldresp = (HTTPHeader)entry.getDataHook ();

	/* if we should have gotten anything but a 2xx or a 304, 
	 * we should act like IMS and INM was not there.
	 */
	HTTPHeader expfail = checkExpectations (in, oldresp);
	if (expfail != null)
	    return expfail;

	String sims = in.getHeader ("If-Modified-Since");
	String sums = in.getHeader ("If-Unmodified-Since");
	Vector vinm = in.getHeaders ("If-None-Match");
	String et = oldresp.getHeader ("Etag");
	boolean etagMatch = false;
	Date ims = null;
	Date ums = null;
	Date dm = null;
	if (sims != null) 
	    ims = HTTPDateParser.getDate (sims);
	if (sums != null)
	    ums = HTTPDateParser.getDate (sums);
	if (ims != null || ums != null) {
	    String lm = oldresp.getHeader ("Last-Modified");
	    if (lm == null)
		return ematch (etagMatch, oldresp);
	    dm = HTTPDateParser.getDate (lm);
	}
	
	if (ums != null && dm.after (ums))
	    return get412 ();	    

	/* Check if we have a match of etags (Weak comparison). */
	if (et != null) {
	    for (int i = 0; i < vinm.size (); i++) {
		String sinm = (String)vinm.get (i);
		if (sinm != null && 
		    (sinm.equals ("*") ||
		     checkWeakEtag (et, sinm)))
		    etagMatch = true;
	    }
	}

	if (sims == null) {
	    /* No IMS, act upon INM only. */
	    return ematch (etagMatch, oldresp);
	} else {
	    /* Here we may or may not have a etagMatch.
	     * Etagmatch and bad(nonexistant/unparsable..) IMS => act on etag
	     */
	    if (ims == null) {
		Proxy.logError (Proxy.INFO, "unparseable date: " + sims + " for URL: " + in.getRequestURI ());
		return ematch (etagMatch, oldresp);
	    }
	    
	    if (dm == null)
		return ematch (etagMatch, oldresp);
	    
	    if (dm.after (ims)) 
		return null;
	    if (vinm.size () < 1)
		return null;
	    return ematch (etagMatch, oldresp);
	}
    }

    private HTTPHeader ematch (boolean etagMatch, HTTPHeader oldresp) {
	if (etagMatch)
	    return get304 (oldresp);
	else 
	    return null;
    }
    
    /** Filtering the response gave us a blocker, so send it.
     */
    private void sendBadResponse (HTTPHeader badresponse, HTTPInputStream contentStream) {
	setKeepalive (false);
	send (badresponse);
	statuscode = badresponse.getStatusCode ();
	try {
	    if (contentStream != null)
		contentStream.close ();
	} catch (IOException e) {
	    // ignore..
	}
    }
    
    private void checkVary (HTTPHeader req, NCacheEntry entry) {
	if (entry == null)
	    return;
	HTTPHeader resp = (HTTPHeader)entry.getDataHook ();
	Vector varies = resp.getHeaders ("Vary");
	for (int i = 0; i < varies.size (); i++) {
	    String vary = (String)varies.get (i);
	    if (vary.equals ("*")) {
		setMayUseCache (false);
		return;
	    } else {
		HTTPHeader origreq = (HTTPHeader)entry.getKey ();
		Vector vals = origreq.getHeaders (vary);
		Vector nvals = req.getHeaders (vary);
		if (vals.size () != nvals.size ()) {
		    setMayUseCache (false);
		    return;
		} else {
		    for (int j = 0; j < vals.size (); j++) {
			String val = (String)vals.get (j);
			int k = nvals.indexOf (val);
			if (k == -1) {
			    setMayUseCache (false);
			    return;
			}
		    }
		}
	    }
	}
    }

    private void checkNoStore (HTTPHeader req, NCacheEntry entry) {
	if (entry == null)
	    return;
	Vector ccs = req.getHeaders ("Cache-Control");
	for (int i = 0; i < ccs.size (); i++) {
	    String cc = (String)ccs.elementAt (i);
	    if (cc.equals ("no-store")) {
		Proxy.getCache ().remove (entry.getKey ());
	    }
	}
    }

    private boolean checkMaxStale (HTTPHeader req, NCacheEntry entry) {
	Vector ccs = req.getHeaders ("Cache-Control");
	for (int i = 0; i < ccs.size (); i++) {
	    String cc = (String)ccs.elementAt (i);
	    cc.trim ();
	    if (cc.equals ("max-stale")) {
		if (entry != null) {
		    HTTPHeader resp = (HTTPHeader)entry.getDataHook ();
		    long maxAge = getCacheControlValue (resp, "max-age=");
		    if (maxAge >= 0) {
			Date now = new Date ();
			long secs = (now.getTime () - entry.getCacheTime ().getTime ()) / 1000;
			long currentAge = secs;
			String age = resp.getHeader ("Age");
			if (age != null)
			    currentAge += Long.parseLong (age);
			if (currentAge > maxAge) {
			    resp.addHeader ("Warning", "110 RabbIT \"Response is stale\"");
			}
		    }
		}
		return true;
	    }
	}
	return false;
    }

    private boolean checkMaxAge (NCacheEntry entry) {
	if (entry == null)
	    return false;
	HTTPHeader resp = (HTTPHeader)entry.getDataHook ();
	return checkMaxAge (resp, entry);
    }

    private boolean checkMaxAge (HTTPHeader req, NCacheEntry entry) {
	if (entry == null)
	    return false;
 	Vector ccs = req.getHeaders ("Cache-Control");
	for (int i = 0; i < ccs.size (); i++) {
	    if (checkMaxAge (req, entry, (String)ccs.elementAt (i)))
	 	return true;
	}
	return false;
    }
    
    private boolean setupRevalidation (HTTPHeader req, NCacheEntry entry) {
	setMayUseCache (false);
	String method = req.getMethod ();
	if (method.equals ("GET") && entry != null) {
	    HTTPHeader resp = (HTTPHeader)entry.getDataHook ();
	    String etag = resp.getHeader ("ETag");
	    String lmod = resp.getHeader ("Last-Modified");
	    if (etag != null) {
		String inm = req.getHeader ("If-None-Match");
		if (inm == null) {
		    req.setHeader ("If-None-Match", etag);
		    return true;
		}
	    } else if (lmod != null) {
		String inm = req.getHeader ("If-Modified-Since");
		if (inm == null) {
		    req.setHeader ("If-Modified-Since", lmod);
		    return true;
		}		
	    } else {
		setMayUseCache (false);
		return false;
	    }
	}
	return false;
    }

    private boolean checkMaxAge (HTTPHeader req, NCacheEntry entry, String cached) {
	if (cached != null) {
	    cached = cached.trim(); 
	    
	    if (cached.startsWith ("max-age=0")) {
		return setupRevalidation (req, entry);
	    } else {
		Date now = new Date ();
		if (checkMaxAge (cached, "max-age=", entry.getCacheTime (), now) ||
		    checkMaxAge (cached, "s-maxage=", entry.getCacheTime (), now)) {
		    setMayUseCache (false);
		    return false;
		}
	    }
	}
	return false;
    }
    
    private boolean checkMaxAge (String cached, String type, Date cachetime, Date now) {
	if (cached.startsWith (type)) {
	    String secs = cached.substring (type.length ());
	    try {
		long l = Long.parseLong (secs) * 1000;
		long ad = now.getTime () - cachetime.getTime ();
		if (ad > l)
		    return true;
	    } catch (NumberFormatException e) {
		Proxy.logError (Proxy.WARN, "Bad number for max-age: '" + cached.substring (8) + "'");		    
	    }
	}
	return false;
    }

    private boolean checkNoCacheHeader (Vector v) {
	for (int i = 0; i < v.size (); i++) {
	    String nc = (String)v.elementAt (i);
	    if (nc.equals ("no-cache")) {
		return true;
	    }
	}
	return false;
    }

    private boolean checkNoCache (HTTPHeader header, NCacheEntry entry) {
	if (entry == null)
	    return false;
	boolean noCache = false;
	// Only check the response header, request headers with no-cache == refetch.
	HTTPHeader resp = (HTTPHeader)entry.getDataHook ();
	noCache = checkNoCacheHeader (resp.getHeaders ("Cache-Control"));
	if (noCache) {
	    return setupRevalidation (header, entry);
	}
	return false;
    }

    private boolean checkQuery (HTTPHeader header, NCacheEntry entry) {
	if (entry == null)
	    return false;
	String uri = header.getRequestURI ();
	int i = uri.indexOf ('?');
	if (i >= 0) {
	    return setupRevalidation (header, entry);
	}
	return false;
    }

    private long getCacheControlValue (HTTPHeader header, String cc) {
	Vector v = header.getHeaders ("Cache-Control");
	for (int i = 0; i < v.size (); i++) {
	    String nc = (String)v.elementAt (i);
	    if (nc.startsWith (cc)) {
		return Long.parseLong (nc.substring (cc.length ()));
	    }
	}
	return -1;
    }

    private boolean checkMinFresh (HTTPHeader header, NCacheEntry entry) {
	if (entry == null)
	    return false;
	long minFresh = getCacheControlValue (header, "min-fresh=");
	if (minFresh == -1)
	    return false;
	long maxAge = getCacheControlValue ((HTTPHeader)entry.getDataHook (), "max-age=");
	if (maxAge == -1)
	    return false;
	long currentAge = (new Date ().getTime () - entry.getCacheTime ().getTime ()) / 1000;
	if ((maxAge - currentAge) < minFresh)
	    return setupRevalidation (header, entry);
	return false;
    }

    private boolean checkRevalidation (HTTPHeader header, NCacheEntry entry) {
	if (entry == null)
	    return false;
	
	HTTPHeader resp = (HTTPHeader)entry.getDataHook ();
	Vector v = resp.getHeaders ("Cache-Control");
	for (int i = 0; i < v.size (); i++) {
	    String nc = (String)v.elementAt (i);
	    if (nc.equals ("must-revalidate") ||
		nc.equals ("proxy-revalidate")) {
		long maxAge = getCacheControlValue ((HTTPHeader)entry.getDataHook (), "max-age=");
		if (maxAge >= 0) {
		    long currentAge = (new Date ().getTime () - entry.getCacheTime ().getTime ()) / 1000;
		    if (currentAge > maxAge) {			
			return setupRevalidation (header, entry);
		    }
		}
	    } else if (nc.startsWith ("s-maxage=")) {
		long sm = Long.parseLong (nc.substring ("s-maxage=".length ()));
		if (sm >= 0) {
		    long currentAge = (new Date ().getTime () - entry.getCacheTime ().getTime ()) / 1000;
		    if (currentAge > sm) {			
			return setupRevalidation (header, entry);
		    }
		}
	    }
	}
	return false;
    }   

    private boolean checkConditional (HTTPHeader header, NCacheEntry entry) {
	return (checkMaxAge (header, entry) || 
		checkNoCache (header, entry) ||
		checkQuery (header, entry) ||
		checkMinFresh (header, entry) || 
		checkRevalidation (header, entry));
    }

    private class RequestHandler {
	public HTTPInputStream contentStream = null;
	public HTTPHeader webheader = null;
	public NCacheEntry entry = null;
	public Class chandler = null;
	public long size = -1;
	public WebConnection wc = null;	
	public Date requestTime = null;
    }

    private void setAge (RequestHandler rh) {
	Date now = new Date ();
	String age = rh.webheader.getHeader ("Age");
	String date = rh.webheader.getHeader ("Date");
	Date dd = HTTPDateParser.getDate (date);
	if (dd == null)
	    dd = now;
	long lage = 0;
	try {
	    if (age != null) {
		lage = Long.parseLong (age);
	    }
	    long dt = Math.max ((now.getTime () - dd.getTime () - Proxy.getOffset ()) / 1000, 0);
	    long correct_age = lage + dt;
	    long correct_recieved_age = Math.max (dt, lage);
	    long corrected_initial_age = correct_recieved_age + dt;
	    if (corrected_initial_age > 0)
		rh.webheader.setHeader ("Age", "" + corrected_initial_age);
	} catch (NumberFormatException e) {
	    // if we cant parse it, we leave the Age header..
	    Proxy.logError (Proxy.WARN, "Bad age: " + age);
	}
    }

    private int nextNonBlank (String s, int start) {
	char c;
	int len = s.length ();
	while (start < len && ((c = s.charAt (start)) == ' ' || c == '\n' || c == '\r' || c == '\t'))
	    start++;
	return start;
    }
    
    private int nextBlank (String s, int start) {
	char c;
	int len = s.length ();
	while (start < len && !((c = s.charAt (start)) == ' ' || c == '\n' || c == '\r' || c == '\t'))
	    start++;
	return start;
    }

    private void removeWarnings (HTTPHeader header, boolean remove1xx) {
	Vector warns = header.getHeaders ("Warning");	
	String rdate = header.getHeader ("Date");
	for (int w = 0; w < warns.size (); w++) {
	    String val = (String)warns.elementAt (w);
	    try {
		StringBuffer sb = new StringBuffer ();
		boolean first = true;
		int start = 0;
		while (start < val.length ()) {
		    int i = nextNonBlank (val, start);
		    i = nextBlank (val, i);
		    String code = val.substring (start, i);
		    int j = nextNonBlank (val, i + 1);
		    j = nextBlank (val, j);
		    String agent = val.substring (i + 1, j);
		    int k = val.indexOf ('"', j);
		    int l = val.indexOf ('"', k + 1);
		    String text = val.substring (k + 1, l);   // StringIndexOutOfBoundsException: -1
		    int c = val.indexOf (',', l);		
		    int m = val.indexOf ('"', l + 1);
		    String date = null;
		    if (((c == -1 && m == -1) || (c < m))) { 
			start = l + 1;
		    } else {
			int n = val.indexOf ('"', m + 1);
			date = val.substring (m + 1, n);
			int c2 = val.indexOf (',', n + 1);
			if (c2 != -1)
			    start = c2;
			else 
			    start = n + 1;
		    }
		    char s;
		    while (start < val.length () && ((s = val.charAt (start)) == ' ' || s == ','))
			start++;
		    
		    if ((date != null && !date.equals (rdate)) ||
			(remove1xx && code.charAt (0) == '1') && !"RabbIT".equals (agent)) {
		    } else {
			if (!first)
			    sb.append (", ");
			sb.append (code + " " + agent + " \"" + text + (date != null ? "\" \"" + date + "\"" : "\"")); 
			first = false;
		    }
		}
		if (sb.length () != 0) 
		    header.setExistingValue (val, sb.toString ());
		else 
		    header.removeValue (val);
	    } catch (StringIndexOutOfBoundsException e) {
		Proxy.logError (Proxy.WARN, "bad warning header: '" + val + "'");
	    }
	}
    }
    
    private void updateWarnings (HTTPHeader header, HTTPHeader webheader) {
	Vector warnings = webheader.getHeaders ("Warning");
	for (int i = 0; i < warnings.size (); i++)
	    header.addHeader ("Warning", (String)warnings.elementAt (i));
    }

    private void updateHeader (RequestHandler rh) {
	if (rh.entry == null)
	    return;
	HTTPHeader cachedheader = (HTTPHeader)rh.entry.getDataHook ();	
	String date = rh.webheader.getHeader ("Date");
	if (date != null)
	    cachedheader.setHeader ("Date", date);
	removeWarnings (cachedheader, true);
	updateWarnings (cachedheader, rh.webheader);	
    }

    private void setupCachedEntry (HTTPHeader header, RequestHandler rh) throws IOException {
	rh.contentStream = new HTTPInputStream (new CacheStream (rh.entry));
	setChunking (false);
	rh.webheader = (HTTPHeader)rh.entry.getDataHook ();
	//rh.webheader.setHeader ("Content-Length", "" + rh.entry.getSize ());
	rh.size = rh.entry.getSize ();
	String age = rh.webheader.getHeader ("Age");
	Date now = new Date ();
	long secs = (now.getTime () - rh.entry.getCacheTime ().getTime ()) / 1000;
	if (age != null) {
	    try {
		long l = Long.parseLong (age);
		secs += l;
	    } catch (NumberFormatException e) {
		Proxy.logError (Proxy.WARN, "bad Age : '" + age + "'");
	    }
	} 
	rh.webheader.setHeader ("Age", "" + secs);
	// do we have a handler for it?
	String ctype = rh.webheader.getHeader ("Content-Type");
	if (ctype != null)
	    rh.chandler = (Class)Proxy.cachehandlers.get (ctype);
	if (rh.chandler == null)
	    rh.chandler = BaseHandler.class; // and simply send, its already filtered.
	removeWarnings (rh.webheader, false);
    }
    
    private void setupWebConnection (HTTPHeader header, RequestHandler rh) throws IOException {
	int attempts = 0;
	do {
	    rh.requestTime = new Date ();
	    rh.wc = Proxy.getWebConnection (header);
	    try {
		if (header.getContentStream () != null) 
		    header.setHeader ("Transfer-Encoding", "chunked");
		rh.wc.writeHTTPHeader (header);		
		HTTPInputStream his = header.getContentStream ();
		if (his != null) {
		    InputStream is = his.getChunkStream ();
		    chunkData (rh.wc, is);
		}
	    } catch (IOException e) {
		// damn, something seems closed....
		// try a new one.
		Proxy.getCounter ().inc ("WebConnections failed on write");
		//e.printStackTrace ();
 		rh.wc = null;
		continue;
	    }

	    try {
		rh.contentStream = rh.wc.getInputStream ();
		if (!header.isDot9Request ()) {
		    char status = '0';
		    do {
			rh.webheader = rh.contentStream.readHTTPHeader ();
			// since we have posted the full request loop while 100 (continue) response.
			if ((status = rh.webheader.getStatusCode ().charAt (0)) == '1') {
			    //if http/1.1
			    if (requestline.endsWith ("1.1"))
				out.writeHTTPHeader (rh.webheader);
			}
		    } while (status == '1');
		    rh.size = rh.wc.dataSize ();
		    if (rh.wc.chunked ()) {
			rh.contentStream = new HTTPInputStream (rh.wc.getChunkStream ());
		    }
		    String responseVersion = rh.webheader.getResponseHTTPVersion ();
		    setAge (rh);
		    removeWarnings (rh.webheader, false);
		    rh.webheader.addHeader ("Via", responseVersion + " RabbIT");
		}
		else 
		    rh.chandler = BaseHandler.class;   // dont know the mime...
	    } catch (IOException e) {
		// damn, something seems closed....
		// try a new one.
		Proxy.getCounter ().inc ("WebConnections failed on read header");
		//e.printStackTrace ();
 		rh.wc = null;
	    }
	    attempts++;
	} while (rh.wc == null && attempts < 5);   // if we cant get a connection in five cancel..
	if (rh.wc == null)
	    throw new IOException ("Unable to connect to URL: " + header.getRequestURI ());

    }

    private boolean partialContent (HTTPHeader request, RequestHandler rh) {
	if (rh.entry == null)
	    return false;
	String method = request.getMethod ();
	if (!method.equals ("GET"))
	    return false;
	HTTPHeader resp = (HTTPHeader)rh.entry.getDataHook ();
	String realLength = resp.getHeader ("RabbIT-Partial");
	return (realLength != null);
    }
    
    private void fillupContent (HTTPHeader request, RequestHandler rh) {
	HTTPHeader resp = (HTTPHeader)rh.entry.getDataHook ();
	String realLength = resp.getHeader ("RabbIT-Partial");
	long l = Long.parseLong (realLength);
	request.setHeader ("Range", "bytes=" + rh.entry.getSize () + "-");
	try {
	    setupWebConnection (request, rh);
	    String status = rh.webheader.getStatusCode ();
	    if (status.equals ("200")) {
		setMayUseCache (false);
		Proxy.releaseWebConnection (rh.wc);	    
		return;
	    } else if (status.equals ("206")) {
		Proxy.logError (Proxy.WARN, "Filling up partial request is not yet implemented:");
		Proxy.releaseWebConnection (rh.wc);	    
		return;
	    }
 	} catch (IOException e) {
	    Proxy.logError (Proxy.WARN, "Error filling up partial request:" + e);
	    return;	    	    
	}
	setMayCache (true);
    }

    private void setMayCacheFromCC (HTTPHeader header, RequestHandler rh) {
	HTTPHeader resp = rh.webheader;
	Vector ccs = resp.getHeaders ("Cache-Control");
	for (int i = 0; i < ccs.size (); i++) {
	    String val = (String)ccs.elementAt (i);
	    if ("public".equals (val) || 
		"must-revalidate".equals (val) ||
		val.startsWith ("s-maxage=")) {
		String auth = header.getHeader ("Authorization");
		if (auth != null) {
		    // TODO this ignores no-store and a few other things...
		    maycache = true;
		    break;
		}
	    }
	}
    }

    /** Handle a request by getting the datastream (from the cache or the web).
     *  After getting the handler for the mimetype, send it.
     * @param header the request made.
     */
    public void handleRequest (HTTPHeader header) {
	RequestHandler rh = new RequestHandler ();
	
	String method = header.getMethod ().trim ();
	if (!method.equals ("GET") && !method.equals ("HEAD")) {
	    Proxy.getCache ().remove (header);
	}
	
	rh.entry = Proxy.getCache ().getEntry (header);
	checkVary (header, rh.entry);
	checkNoStore (header, rh.entry);
	if (!checkMaxStale (header, rh.entry) && checkMaxAge (rh.entry))
	    setMayUseCache (false);
	
	boolean conditional = checkConditional (header, rh.entry); //checkMaxAge (header, rh.entry);
	if (partialContent (header, rh)) {
	    fillupContent (header, rh);
	}

	if (getMayUseCache ()) {
	    // in cache?
	    if (rh.entry != null) {
		Proxy.getCounter ().inc ("Cache hits");
		setKeepalive (true);
		HTTPHeader resp = checkIfMatch (header, rh.entry);
		if (resp == null)
		    resp = is304 (header, rh.entry);
		if (resp != null) {
		    send (resp);
		    statuscode = resp.getStatusCode ();
		    sendBadResponse (resp, rh.contentStream);
		    return;
		}
		setMayCache (false);
		try {		    
		    setupCachedEntry (header, rh);			
		} catch (FileNotFoundException e) {
		    rh.contentStream = null;         // ignore sorta, to pull resource from the web.
		    rh.entry = null;
		} catch (IOException e) {
		    rh.contentStream = null;
		    rh.entry = null;
		}
	    }
	}
	
	if (rh.contentStream == null) {
	    // no, not in cache.
	    try {
		setupWebConnection (header, rh);
		setMayCacheFromCC (header, rh);
	    } catch (IOException e) {
		Proxy.logError (Proxy.DEBUG, "strange error: " + e.toString ());
		doError (400, e);
		setKeepalive (false);
		return;
	    }
	}
	
	// and now we filter the response header if any.
	if (!header.isDot9Request ()) {
	    String status = rh.webheader.getStatusCode ().trim ();
	    rh.entry = Proxy.getCache ().getEntry (header);
	    checkStaleCache (header, rh.webheader);	    
	    removeOtherStaleCaches (header, rh.webheader);
	    if (status.equals ("304"))
		updateHeader (rh);
	    HTTPHeader badresponse = checkExpectations (header, rh.webheader);
	    if (badresponse == null)
		badresponse = filterHTTP (Proxy.httpoutfilters, rh.webheader);
	    if (badresponse != null) {
		sendBadResponse (badresponse, rh.contentStream);
		return;
	    }
	    if (conditional && (rh.entry = Proxy.getCache ().getEntry (header)) != null && 
		status.equals ("304")) {
		HTTPHeader cachedheader = (HTTPHeader)rh.entry.getDataHook ();
		Proxy.releaseWebConnection (rh.wc);
		if (checkWeakEtag (cachedheader, rh.webheader)) {
		    updateHeader (rh);
		    setMayCache (false);
		    try {
			setupCachedEntry (header, rh);
		    } catch (FileNotFoundException e) {
			Proxy.logError (Proxy.WARN, "Conditional request could not find the cached file (" + 
					header.getRequestURI () + ", : " + e);
		    } catch (IOException e) {
			Proxy.logError (Proxy.WARN, "Conditional request got IOException (" + 
					header.getRequestURI () + ",: " + e); 
		    }
		} else {
		    // retry... 
		    header.removeHeader ("If-None-Match");
		    Proxy.getCache ().remove (header);
		    handleRequest (header);
		    return;
		}
	    }
	}

	// ok get the handler for it.
	if (rh.chandler == null) {
	    if (getMayFilter () && 
		rh.webheader != null && rh.webheader.getHeader ("Content-Type") != null)
		rh.chandler = (Class)Proxy.handlers.get (rh.webheader.getHeader ("Content-Type"));
	    if (rh.chandler == null)               // still null
		rh.chandler = BaseHandler.class;   // fallback...
	}   
	Handler handler = getNewInstance (rh.chandler, 
					  header, 
					  rh.webheader, 
					  rh.contentStream, 
					  out, 
					  getMayCache (), 
					  getMayFilter (), 
					  rh.size);           // actual size of data if known (-1 = unknown).
	
	if (handler == null) {
	    doError (500, "Something fishy with that handler....");
	    setKeepalive (false);
	} else {
	    try {
		if (chunk) {
		    rh.webheader.removeHeader ("Content-Length");
		    rh.webheader.setHeader ("Transfer-Encoding", "chunked");
		} else {
		    if (getKeepalive ()) {
			rh.webheader.setHeader ("Proxy-Connection", "Keep-Alive");
			rh.webheader.setHeader ("Connection", "Keep-Alive");
		    } else {
			rh.webheader.setHeader ("Proxy-Connection", "close");
			rh.webheader.setHeader ("Connection", "close");
		    }
		}
		
		// HTTP/0.9 does not support HEAD, so rh.webheader should be valid.
		if (header.isHeadOnlyRequest ())   
		    out.writeHTTPHeader (rh.webheader);
		else 
		    handler.handle ();
		if (chunk)
		    client.finish ();
		
		if (rh.wc != null) {
		    Proxy.releaseWebConnection (rh.wc);		    
		} else if (rh.entry != null) {
		    rh.contentStream.close ();
		}
	    } catch (IOException e){
		Proxy.logError (Proxy.WARN, "Error writing request:" + e);
		setKeepalive (false);
		return;	    
	    }
	}	
	/* No dont close, let the ConnectionHandler handle that... */
    }

    private void chunkData (WebConnection wc, InputStream is) throws IOException {
	byte[] buf = new byte[1024];
	int read; 
	HTTPOutputStream os = wc.getOutputStream ();
	while ((read = is.read (buf)) > 0)
	    os.write (buf, 0, read);
	os.finish ();
    }

    private void checkStaleCache (HTTPHeader header, HTTPHeader webheader) {
	NCacheEntry entry = Proxy.getCache ().getEntry (header);
	if (entry == null)
	    return;
	if (webheader.getStatusCode ().trim ().equals ("304"))
	    return;
	HTTPHeader cachedwebheader = (HTTPHeader)entry.getDataHook ();
	checkStaleHeader (header, webheader, cachedwebheader, "Content-Length");
	checkStaleHeader (header, webheader, cachedwebheader, "Content-MD5");
	checkStaleHeader (header, webheader, cachedwebheader, "ETag");
	checkStaleHeader (header, webheader, cachedwebheader, "Last-Modified");
    }

    private void removeCaches (HTTPHeader header, HTTPHeader webheader, String type) {
	String loc = webheader.getHeader (type);
	if (loc == null)
	    return;
	try {
	    URL u = new URL (header.getRequestURI ());
	    URL u2 = new URL (u, loc);
	    HTTPHeader h = new HTTPHeader ();
	    h.setRequestURI (u2.toString ());
	    Proxy.getCache ().remove (h);
	} catch (MalformedURLException e) {
	    Proxy.logError (Proxy.WARN, "Trying to delete cached object with bad url (" + 
			    header.getRequestURI () + ", " + loc + ": " + e);
	}
    }

    private void removeCaches (HTTPHeader header, HTTPHeader webheader) {
	removeCaches (header, webheader, "Location");
	removeCaches (header, webheader, "Content-Location");
    }

    private void removeOtherStaleCaches (HTTPHeader header, HTTPHeader webheader) {
	String method = header.getMethod ();
	String status = webheader.getStatusCode ();
	if ((method.equals ("PUT") || method.equals ("POST")) && status.equals ("201")) {
	    removeCaches (header, webheader);
	} else if (method.equals ("DELETE") && status.equals ("200")) {
	    removeCaches (header, webheader);
	}
    }

    private void checkStaleHeader (HTTPHeader header, HTTPHeader webheader, 
				   HTTPHeader cachedwebheader, String str) {
	String cln = webheader.getHeader (str);
	String clo = cachedwebheader.getHeader (str);
	if (clo != null) {
	    if (!clo.equals (cln)) {
		Proxy.getCache ().remove (header);
		return;
	    }
	} else { 
	    if (cln != null) {
		Proxy.getCache ().remove (header); 
		return; 
	    }
	}
    }

    private HTTPHeader checkExpectations (HTTPHeader header, HTTPHeader webheader) {
	String exp = header.getHeader ("Expect");
	if (exp == null)
	    return null;
	if (exp.equals ("100-continue")) {
	    String status = webheader.getStatusCode ();
	    if (status.equals ("200") || status.equals ("304"))
		return null;
	    return get417 (exp);
	}
	StringTokenizer st = new StringTokenizer (exp, ";");
	while (st.hasMoreTokens ()) {
	    String e = st.nextToken ();
	    int i = e.indexOf ('=');
	    if (i == -1 || i == e.length () -1) 
		return get417 (e);
	    String type = e.substring (0, i);
	    String value = e.substring (i + 1);
	    if (type.equals ("expect")) {
		String h = webheader.getHeader (value);
		if (h == null)
		    return get417 ("No expected header found");
	    }
	}
	
	return get417 (exp);
    }

    /** Handle a meta page.
     * @param header the request being made.
     */
    public void handleMeta (HTTPHeader header) {
	Proxy.getCounter ().inc ("Meta pages requested");
	URL url = null;
	try {
	    url = new URL (header.getRequestURI ());
	} catch (MalformedURLException e) {
	    // this should not happen since HTTPBaseHandler managed to do it...
	}
	String file = url.getFile ().substring (1);  // remove initial '/'
	if (file.length () == 0)
	    file = "FileSender/";
	
	int index = -1;
	String args = "";
	if ((index = file.indexOf ("?")) >= 0) {
	    args = file.substring (index + 1);
	    file = file.substring (0, index);
	}
	Properties htab = splitArgs (args);
 	if ((index = file.indexOf ("/")) >= 0) {
	    String fc = file.substring (index + 1);
	    file = file.substring (0, index);
	    htab.put ("argstring", fc);
	}
	String error = null;
	try {
	    if (file.indexOf (".") < 0)
		file = "rabbit.meta." + file;
	    
	    Class cls = Class.forName (file);
	    MetaHandler mh = (MetaHandler)metaHandlers.get (cls);
	    if (mh == null) {
		mh = (MetaHandler)cls.newInstance ();
		metaHandlers.put (cls, mh);
	    }
	    mh.handle (in, out, header, htab, this);
	    Proxy.getCounter ().inc ("Meta pages handled");	
	    // Now take care of every error...
	} catch (NoSuchMethodError e) {
	    error = "Given metahandler doesnt have a public no-arg constructor:" + file + ", " + e;
	} catch (ClassCastException e) {
	    error = "Given metapage is not a MetaHandler:" + file + ", " + e;
	} catch (ClassNotFoundException e) {
	    error = "Couldnt find class:" + file + ", " + e;
	} catch (InstantiationException e) {
	    error = "Couldnt instantiate metahandler:" + file + ", " + e;
	} catch (IllegalAccessException e) {
	    error = "Que? metahandler access violation?:" + file + ", " + e;
	} catch (IllegalArgumentException e) {
	    error = "Strange name of metapage?:" + file + ", " + e;
	} 
	if (error != null) {
	    Proxy.logError (Proxy.WARN, error);
	    doError (400, error);
	    return;
	}
    }


    /** Check to see if this header is an SSL header.
     * @param header the header to check.
     * @return true if this header is an SSL request.
     */
    public boolean isSSLRequest (HTTPHeader header) {
	return header.getMethod ().equals ("CONNECT");
    }

    /** Handle the SSL request
     * @param header the Connect request header.
     */
    protected void handleSSL (HTTPHeader header) {
	status = "handling SSL connection";

	String uri = header.getRequestURI ();
	header.setRequestURI ("http://" + uri);  // java 1.1 does not know about https.
	try {
	    // handle proxy connections here.
	    WebConnection wc = Proxy.getWebConnection (header);
	    
	    HTTPHeader reply = new HTTPHeader ();
	    reply.setStatusLine ("HTTP/1.0 200 Connection established");
	    reply.setHeader ("Proxy-agent", Proxy.getServerIdentity ());
	    out.writeHTTPHeader (reply);
	    
	    OutputStream server = wc.getOutputStream ();
	    InputStream resp = wc.getInputStream ();
	    byte[] buf = new byte[1024];
	    int read;
	    CopyThread cp = new CopyThread (resp, out);
	    while ((read = in.read(buf)) > 0) {
		server.write (buf, 0, read);
	    }
	} catch (SocketException e) {
	    // ignore since this will most probably be the closing of the connection...
	} catch (Exception e) {
	    e.printStackTrace ();
	    doError (400, e);
	}
    }

    /** Are we allowed to proxy ssl-type connections ?
     * @return true if we allow the CONNECT &lt;port&gt; command.
     */	
    public boolean sslIsAllowed (HTTPHeader header) {
	String hp = header.getRequestURI ();
	int c = hp.indexOf (':');
	Integer port = new Integer (443);
	if (c >= 0) {
	    try {
		port = new Integer (hp.substring (c+1));
	    } catch (Exception e) {
		// strange request, dont allow it.
		return false;
	    }
	}
	if (Proxy.proxySSL == false)
	    return false;
	if (Proxy.proxySSL == true && Proxy.sslports == null)
	    return true;
	for (int i = 0; i < Proxy.sslports.size (); i++) {
	    if (port.equals (Proxy.sslports.elementAt (i)))
		return true;
	}
	return false;
    }
    
    /** splits the CGI-paramsstring into variables and values.
     *  put these values into a hashtable for easy retrival
     * @param params the CGI-querystring.
     * @return a hastable with type->value maps for the CGI-querystring
     */
    protected Properties splitArgs (String params) {
	Properties htab = new Properties ();
	StringTokenizer st = new StringTokenizer (params, "=&", true);
	String key = null;
	while (st.hasMoreTokens ()) {
	    String next = st.nextToken ();
	    if (next.equals ("=")) {
		// nah..
	    } else if (next.equals ("&")) {
		if (key != null) {
		    htab.put (key, "");
		    key = null;
		}
	    } else if (key == null) {
		key = next;
	    } else {
		htab.put (key, Coder.URLdecode (next));
		key = null;
	    }
	}
	return htab;
    }
    
    /** Send an error (400 Bad Request) to the client.
     * @param statuscode the status code of the error.
     * @param message the error message to tell the client.
     */
    public void doError (int statuscode, String message) {
	this.statuscode = "" + statuscode;
	HTTPHeader header = getHeader ("HTTP/1.0 400 Bad Request");
	StringBuffer error = new StringBuffer (HTMLPage.getPageHeader ("400 Bad Request") +
					       "Unable to handle request:<br><b>" + 
					       message +
					       "</b></body></html>\n");
	header.setContent (error.toString ());
	send (header);
    }

    /** Send an error (400 Bad Request) to the client.
     * @param statuscode the status code of the error.
     * @param message the error message to tell the client.
     */
    public void doError (int statuscode, Exception e) {
	ByteArrayOutputStream bos = new ByteArrayOutputStream ();
	PrintWriter ps = new PrintWriter (new OutputStreamWriter (bos));
	e.printStackTrace (ps);
	String message = bos.toString ();
	this.statuscode = "" + statuscode;
	HTTPHeader header = getHeader ("HTTP/1.0 400 Bad Request");
	StringBuffer error = new StringBuffer (HTMLPage.getPageHeader ("400 Bad Request") +
					       "Unable to handle request:<br><b>" + 
					       e.getMessage () +
					       "</b><br><xmp>" + message + "</xmp></body></html>\n");
	header.setContent (error.toString ());
	send (header);
    }

    /** Get the status of this Connection
     * @return the current status.
     */
    public String getStatus () {
	return status;
    }
    
    /** Get the time this Connection was started.
     * @return the time this Connection started serving requests.
     */
    public Date getStarted () {
	return started;
    }
    
    /** Set the user name of the client.
     * @param username the username of the client.
     */
    public void setUserName (String username) {
	this.username = username;
    }

    /** Get the username of the client.
     * @return the current username of the client.
     */
    public String getUserName () {
	return username;
    }

    /** Set the password of the client.
     * @param password the password of the client.
     */
    public void setPassWord (String password) {
	this.password = password;
    }
    
    /** Get the password of the client.
     * @return the current password of the client.
     */
    public String getPassWord () {
	return password;
    }
    
    /** Set keepalive to a new value. Note that keepalive can only be
     *	promoted down. 
     * @param keepalive the new keepalive value.
     */
    public void setKeepalive (boolean keepalive) {
	this.keepalive = (this.keepalive && keepalive);
    }
    
    /** Get the keepalive value.
     * @return true if keepalive should be done, false otherwise.
     */
    public boolean getKeepalive () {
	return keepalive;
    }    

    /** Set the chunking option.
     * @param b if true this connection should use chunking.
     */
    public void setChunking (boolean b) {
	chunk = b;
    }

    /** Get the chunking option.
     * @return if this connection is using chunking.
     */
    public boolean getChunking () {
	return chunk;
    }
    
    /** Set the state of this request.
     * @param meta true if this request is a metapage request, false otherwise.
     */
    public void setMeta (boolean meta) {
	this.meta = meta;
    }
    
    /** Get the state of this request.
     * @return true if this is a metapage request, false otherwise.
     */
    public boolean getMeta () {
	return meta;
    }

    /** Set the state of this request. This can only be promoted down..
     * @param usecache true if we may use the cache for this request, false otherwise.
     */
    public void setMayUseCache (boolean usecache) {
	mayusecache = mayusecache && usecache;
    }

    /** Get the state of this request.
     * @return true if we may use the cache for this request, false otherwise.
     */
    public boolean getMayUseCache () {
	return mayusecache;
    }
    
    /** Set the state of this request. This can only be promoted down.
     * @param cacheAllowed true if we may cache the response, false otherwise.
     */
    public void setMayCache (boolean cacheAllowed) {
	maycache = cacheAllowed && maycache;
    }

    /** Get the state of this request.
     * @return true if we may cache the response, false otherwise.
     */
    public boolean getMayCache () {
	return maycache;
    }

    /** Get the state of this request. This can only be promoted down.
     * @param filterAllowed true if we may filter the response, false otherwise.
     */
    public void setMayFilter (boolean filterAllowed) {
	mayfilter = filterAllowed && mayfilter;
    }
    
    /** Get the state of the request.
     * @return true if we may filter the response, false otherwise.
     */
    public boolean getMayFilter () {
	return mayfilter;
    }

    /** Get the Socket that is being served.
     * @return the requesting Socket.
     */
    public Socket getSocket () {
	return socket;
    }

    /** Get the current request line
     * @return the request being handled.
     */
    public String getRequestLine () {
	return requestline;
    }
    
    /** Get the status code of the request.
     * @return the current status code.
     */
    public String getStatusCode () {
	return statuscode;
    }

    /** Get the extra information for this request.
     * @return the extra information if set.
     */
    public String getExtraInfo () {
	return extrainfo;
    }

    /** Set the extra information for this request.
     * @param exinfo the new extra information to set.
     */
    public void setExtraInfo (String exinfo) {
	extrainfo = exinfo;
    }
    
    /** Set the content length of the response.
     * @param cl the new content length.
     */
    public void setContentLength (String cl) {
	contentlength = cl;
    }
    
    /** Get the content length of the response.
     * @return the content length of the response if set.
     */
    public String getContentLength () {
	return contentlength;
    }
    
    /** Get a new HTTPHeader. This is the same as 
     * getHeader ("HTTP/1.0 200 OK");
     * @return a new HTTPHeader.
     */
    public static HTTPHeader getHeader () {
	return getHeader ("HTTP/1.0 200 OK");
    }
    
    /** Get a new HTTPHeader initialized with some data.
     * @param statusline the statusline of the response.
     * @return a new HTTPHeader.
     */
    public static HTTPHeader getHeader (String statusLine) {
	HTTPHeader ret = new HTTPHeader ();
	ret.setStatusLine (statusLine);
	ret.setHeader ("Server", Proxy.getServerIdentity ());
	ret.setHeader ("Content-type", "text/html");
	ret.setHeader ("Pragma", "no-cache");
	return ret;    
    }

    /** Get a 200 Ok header
     * @return a 200 HTTPHeader .
     */
    public HTTPHeader get200 () {
	HTTPHeader header = getHeader ("HTTP/1.1 200 Ok");  
	header.setHeader ("Date", HTTPDateParser.getDateString (new Date()));
	return header;
    }
    

    /** Get a 304 Not Modified header for the given old header
     * @param oldresp the cached header.
     * @return a 304 HTTPHeader .
     */
    public HTTPHeader get304 (HTTPHeader oldresp) {
	HTTPHeader header = getHeader ("HTTP/1.1 304 Not Modified");  
	String ds = oldresp.getHeader ("Date");
	if (ds != null)
	    header.setHeader ("Date", HTTPDateParser.getDateString (new Date()));
	String cl = oldresp.getHeader ("Content-Location");
	if (cl != null)
	    header.setHeader ("Content-Location", cl);
	String es = oldresp.getHeader ("ETag");
	if (es != null)
	    header.setHeader ("ETag", es);
	return header;
    }

    /** Get a 400 Bad Request header for the given exception.
     * @param exception the Exception handled.
     * @return a HTTPHeader for the exception.
     */
    public HTTPHeader get400 (Exception exception) {
	// in most cases we should have a header out already, but to be sure...
	HTTPHeader header = getHeader ("HTTP/1.0 400 Bad Request ");  
	StringBuffer accreq = new StringBuffer (HTMLPage.getPageHeader ("400 Bad Request") +
						"Unable to handle request:<br><b><XMP>\n" + 
						exception +
						"</XMP></b></body></html>\n");
	header.setContent (accreq.toString ());
	return header;
    }

    /** Get a 403 Forbidden header.
     * @return a HTTPHeader.
     */
    public HTTPHeader get403 () {
	// in most cases we should have a header out already, but to be sure...
	HTTPHeader header = getHeader ("HTTP/1.0 403 Forbidden");  
	StringBuffer accreq = new StringBuffer (HTMLPage.getPageHeader ("403 Forbidden") +
						"That is forbidden</body></html>");
	header.setContent (accreq.toString ());
	return header;
    }

    /** Get a 407 Proxy Authentication Required for the given realm and url.
     * @param realm the realm that requires auth.
     * @param url the URL of the request made.
     * @return a suitable HTTPHeader.
     */
    public HTTPHeader get407 (String realm, URL url) {
	HTTPHeader header = getHeader ("HTTP/1.0 407 Proxy Authentication Required");
	header.setHeader ("Proxy-Authenticate", "Basic realm=\"" + realm + "\"");
	StringBuffer accreq = new StringBuffer (HTMLPage.getPageHeader ("407 Proxy Authentication Required") +
						"access to: <b>" + url + " </b><br>requires some authentication\n" +
						"</body></html>\n");
	header.setContent (accreq.toString ());
	return header;
    }

    /** Get a 412 Precondition Failed header.
     * @param realm the realm that requires auth.
     * @param url the URL of the request made.
     * @return a suitable HTTPHeader.
     */
    public HTTPHeader get412 () {
	HTTPHeader header = getHeader ("HTTP/1.0 412 Precondition Failed");
	StringBuffer accreq = new StringBuffer (HTMLPage.getPageHeader ("412 Precondition Failed") +
						"</body></html>\n");
	header.setContent (accreq.toString ());
	return header;
    }

    /** Get a 417 Expectation Failed header.
     * @param realm the realm that requires auth.
     * @param url the URL of the request made.
     * @return a suitable HTTPHeader.
     */
    public HTTPHeader get417 (String expectation) {
	HTTPHeader header = getHeader ("HTTP/1.0 417 Expectation failed");
	StringBuffer accreq = new StringBuffer (HTMLPage.getPageHeader ("417 Expectation failed") +
						"RabbIT does not handle the '" + expectation + 
						"' kind of expectations yet.</b>\n" +
						"</body></html>\n");
	header.setContent (accreq.toString ());
	return header;
    }


    /** Get a 500 Internal Server Error header for the given exception.
     * @param exception the Exception made.
     * @return a suitable HTTPHeader.
     */
    public HTTPHeader get500 (Throwable exception) {
	// in most cases we should have a header out already, but to be sure...
	// normally this only thrashes the page... Too bad.
	HTTPHeader header = getHeader ("HTTP/1.0 500 Internal Server Error ");  
	StringWriter sw = new StringWriter ();
	PrintWriter sos = new PrintWriter (sw);
	exception.printStackTrace (sos);
	
	Properties sysprop = System.getProperties ();
	
	StringBuffer accreq = new StringBuffer (HTMLPage.getPageHeader ("500 Internal Server Error") +
						"You have found a bug in RabbIT please report this(together with the URL you tried to visit) to the " +
						"<a href=\"http://www.nada.kth.se/projects/prup98/web_proxy/\" target = \"_top\">RabbIT</a> crew.<br><br>\n" +
						"<font size = 4>Connection status</font><br><hr noshade>\n" + 
						"status: " + status  + "<br>\n" + 
						"started: " + started + "<br>\n" +
						"keepalive: " + keepalive + "<br>\n" + 
						"meta: " + meta + "<br>\n" +
						"mayusecache: " + mayusecache + "<br>\n" + 
						"maycache: " + maycache + "<br>\n" + 
						"mayfilter: " + mayfilter + "<br>\n"+ 
						"requestline: " + requestline + "<br>\n" + 
						"statuscode: " + statuscode + "<br>\n" + 
						"extrainfo: " + extrainfo + "<br>\n" +						
						"contentlength: " + contentlength + "<br>\n<br>\n" +
						"<font size = 4>Proxy status</font><br>\n<hr noshade>\n" +
						"proxy version: " + Proxy.VERSION + "<br>\n" + 
						"proxy identity: " + Proxy.getServerIdentity () + "<br>\n" +
						"server host: " + Proxy.getHost () + "<br>\n" +
						"server port: " + Proxy.getPort () + "<br>\n" +
						"accessfilters: " + Proxy.config.getProperty ("Filters", "accessfilters") + 
						"<br>\n" + 
						"httpinfilters: " + Proxy.config.getProperty ("Filters", "httpinfilters") + 
						"<br>\n" + 
						"httpoutfilters:" + Proxy.config.getProperty ("Filters", "httpoutfilters") +
						"<br>\n<br>\n" + 					       
						"<font size = 4>System properties</font><br>\n<hr noshade>\n" + 
						"java.version: " + sysprop.getProperty ("java.version") + "<br>\n" +
						"java.vendor: " + sysprop.getProperty ("java.vendor") + "<br>\n" +
						"os.name: " + sysprop.getProperty ("os.name") + "<br>\n" +
						"os.version: " + sysprop.getProperty ("os.version") + "<br>\n" +
						"os.arch: " + sysprop.getProperty ("os.arch") + "<br>\n" +
						"error is:<BR><XMP>\n" + sw + 
						"</XMP><br>" + 
						"<hr noshade>\n" +
						"</body></html>\n");
	header.setContent (accreq.toString ());
	return header;
    }
}

