package rabbit.proxy;

// $Id: Proxy.java,v 1.28 2002/09/14 12:30:35 robo Exp $

import java.io.*;
import java.net.*;
import java.util.*;
import java.lang.reflect.*;
import java.text.*;

import rabbit.cache.*;
import rabbit.http.*;
import rabbit.io.*;
import rabbit.util.*;

/** This is the central dispatcher for RabbIT.
 *
 *  Basically it sits in a loop accepting connection and for each
 *  connection creating a handler for it.
 *  
 *  Also handles logging so it goes out nicely formatted (and synchronized).
 *
 */
public class Proxy extends Thread 
{
    /** Current version */
    public static final String VERSION = "RabbIT proxy version 2.0.20";
    
    /** The standard configuration file */
    protected static String CONFIG = "conf/rabbit.conf";
    
    /** Show all messages in the log. */
    public static final int DEBUG = 0;
    /** Show all normal messages and higher. */
    public static final int ALL   = 5;
    /** Show information messages and higher. */
    public static final int INFO  = 10;
    /** Show warnings and higer. */
    public static final int WARN  = 15;
    /** Show important messages and above. */
    public static final int MSG   = 20;
    /** Show error messages and higer. */
    public static final int ERROR = 25;  
    /** Show only fatal messages. */
    public static final int FATAL = 30;
    
    /** Maximum number of concurrent connections */
    private static int maxconnections = 50;

    /** The current connections */
    private static Vector convec = new Vector ();

    /** The log of things */
    private static Counter log = new Counter ();

    /** The config filename for this session */
    private static String configfile = CONFIG;
    
    /** The current configuration */
    public static Config config;
    
    /** Output for errors */
    private static PrintWriter errorlog = new PrintWriter (System.err, true);

    /** Output for accesses */
    private static PrintWriter accesslog = new PrintWriter (System.out, true);

    /** The current error log level. */
    private static int loglevel = MSG;

    /** The time the proxy started */
    private static Date started;

    /** The local adress of the proxy. */
    private static InetAddress localhost;
    /** The local time zone the proxy is running in. */
    private static TimeZone tz;
    /** The distance to GMT in milis. */
    private static long offset;    
    /** The port the proxy is using. */
    private static int port = -1;
    /** The serversocket the proxy is using. */
    private static ServerSocket ss = null;

    /** Adress of connected proxy. */
    private static InetAddress proxy = null;
    /** Port of the connected proxy. */
    private static int proxyport = -1;

    /** The cache-handler */
    private static NCache cache = new NCache ();

    /** The connection handler */
    private static ConnectionHandler conhandler = new ConnectionHandler ();

    /** The current running proxy */
    private static Proxy instance;

    /** The identity of this server. */
    private static String serverIdentity = VERSION;

    /** The format we write dates on. */
    private static SimpleDateFormat sdf = new SimpleDateFormat ("dd/MMM/yyyy:HH:mm:ss 'GMT'");

    /** Is the proxy listening. */
    private static boolean accepting = false;
    
    /** the hash tables with type=>Class mapping */
    protected static Hashtable handlers = new Hashtable ();

    /** the hash tables with type=>Class mapping for cache handling. */
    protected static Hashtable cachehandlers = new Hashtable ();

    /** the filters, a vector of classes (in given order) */
    protected static Vector accessfilters = new Vector ();    
    /** the filters, a vector of classes (in given order) */
    protected static Vector httpinfilters = new Vector ();
    /** the filters, a vector of classes (in given order) */
    protected static Vector httpoutfilters = new Vector ();

    /** Are we allowed to proxy ssl? */
    protected static boolean proxySSL = false;
    /** The Vector of acceptable ssl-ports. */
    protected static Vector sslports = null;

    private static class ConnectionFactory implements RestartableThreadFactory {
	public RestartableThread createThread () {
	    return new Connection ();
	}
    }

    private static RestartableThreadFactory tf = new ConnectionFactory ();
    private static ThreadPool tp = new ThreadPool (tf, maxconnections);
    
    /** Start a proxy. 
     *  Parse flags and read the config, then starts the proxy.
     * @param args the command-line flags given.
     */ 
    public static void main (String[] args) {
	// handle the arguments given...
	int i = 0;
	while (args.length > i) {
	    
	    // use custom ini-file?
	    if (args[i].equals ("-f") || 
		args[i].equals ("--file")) {
		i++;
		if (args.length > i) {
		    configfile = args[i];
		}
		else {
		    logError (FATAL,"No config file specified");
		    System.exit (-1);
		}
	    }
	    
	    // version ?
	    else if (args[i].equals ("-v") || 
		     args[i].equals ("--version")) {
		System.out.println (VERSION);
		System.exit (0);
	    }	

	    // user wants help??
	    else if (args[i].equals ("-?") || 
		     args[i].equals("-h") || 
		     args[i].equals ("--help")) {
		printHelp();
		System.exit (0);
	    } 
	    else {
		System.out.println ("Unsupported argument " + args[i]);	 
	    }
	    i++;
	}
	instance = new Proxy ();	
    }

    /** Constructor, only available from here so we wont have multiple
     *	proxies. 
     */    
    private Proxy () {
	super (VERSION);

	try {
	    localhost = InetAddress.getLocalHost ();
	} catch (UnknownHostException e) {
	    logError (FATAL, "I couldnt find the hostname of this machine, exiting.");
	    System.exit (-1);
	}

	tz = sdf.getTimeZone ();
	GregorianCalendar gc = new GregorianCalendar ();
	gc.setTime (new Date ());
	offset = tz.getOffset (gc.get (Calendar.ERA),
			       gc.get (Calendar.YEAR),
			       gc.get (Calendar.MONTH),
			       gc.get (Calendar.DAY_OF_MONTH),
			       gc.get (Calendar.DAY_OF_WEEK),
			       gc.get (Calendar.MILLISECOND));	
	loadConfig ();
	
	started = new Date ();
	start ();	
    }

    /** Get the time zone the proxy is running in.
     *  This is neccessary to give dates in GMT.
     */
    public TimeZone getTimeZone () {
	return tz;
    }
    
    /** Get the offset in milis from GMT 
     *  NOTE! the offset is only calculated at startup.
     *  This means that during DST changes this may be invalid for a 
     *  day or two. Restart RabbIT dayly during that time to get it right
     *  _IF_ you feel it is a problem.
     */
    public static long getOffset () {
	return offset;
    }
    
    /** Get a Connection that is ready to service.
     */
    private Connection getConnection () {
	return (Connection)tp.getThread ();
    }

    /** Get the connection handler.
     */
    public static ConnectionHandler getConnectionHandler () {
	return conhandler;
    }

    /** Get a WebConnection.
     */
    public static WebConnection getWebConnection (HTTPHeader header) throws IOException {
	return conhandler.getConnection (header);
    }
    
    /** Release a WebConnection so that it may be reused if possible.
     * @param wc the WebConnection to release.
     */
    public static void releaseWebConnection (WebConnection wc) {
	conhandler.releaseConnection (wc);
    }

    /** while we can, accept new sockets, creating a handler for each
     *  and also save connections for statistics.
     */
    public void run () {
	logError (MSG, "Started");
	while (true) {
	    Socket s = null;
	    if (!accepting) {
		try {
		    sleep (10 * 1000);
		} catch (InterruptedException e) {
		}
		continue;
	    }
	    
	    try {
		Connection con = getConnection ();
		s = ss.accept ();
		s.setSoTimeout (120*1000);             // 2 min timeout on read, We should be "near" the client.
		log.inc ("Socket accepts");
		convec.addElement (con);
		con.setSocket (s);
	    } catch (SocketException e) {
		Proxy.logError (WARN, "Socketoptions failed?: " + e);
		try {
		    if (s != null)
			s.close();
		} catch (IOException ee) {}
	    } catch (IOException e) {
		Proxy.logError (ERROR, "Failed to accept, trying to restart serversocket.");
		closeSocket ();
		openSocket ();
	    } catch (Exception e) {
		Proxy.logError (ERROR, "Unknown error: " + e);
	    }
	}
    }

    /** Open a socket on the specified port 
     *  also make the proxy continue accepting connections.
     */
    protected static void openSocket () {
	int tport = Integer.parseInt (config.getProperty (Proxy.class.getName (), "port", "9666").trim ());
	if (tport != port) {
	    try {
		port = tport;
		accepting = false;
		closeSocket ();
		ss = new ServerSocket (port);
		accepting = true;
	    }
	    catch (IOException e) {
		logError (FATAL, "Failed to open serversocket on port " + port);
		System.exit (-1);
	    }
	}
    }
    
    /** Closes the serversocket and makes the proxy stop listening for
     *	connections. 
     */
    protected static void closeSocket () {
	try {
	    if (ss != null) {
		ss.close ();
		ss = null;
	    }
	    accepting = false;
	} catch (IOException e) {
	    logError (FATAL, "Failed to close serversocket on port " + port);
	    System.exit (-1);
	}
    }

    /** print out the helptext to the user.
     */
    public static void printHelp () {
	try {
	    byte[] b = new byte[4096];
	    int i;
	    InputStream f = new FileInputStream ("Help.txt");
	    while ((i = f.read(b)) > 0) 
		System.out.write (b, 0, i);
	    f.close ();
	} catch (IOException e) {
	    logError (WARN, "Couldnt read helptext" + e);
	}
    }

    /** Read the configuration in and set up a lot of things...
     * @param filename the file to read the Config from 
     */
    private void loadConfig () {
	// load the config...
	try {
	    config = new Config (configfile);
	} catch (IOException e) {
	    logError (FATAL, "Could not load the configuration file: '" + configfile + "' exiting");
	    System.exit (-1);
	}	
	setup (config);
    }

    /** Flush and close the logfile given.
     * @param w the logfile to close.
     */
    private void closeLog (Writer w) {
	if (w != null) {
	    try {
		w.flush ();
		w.close ();
	    } catch (IOException e) {
		e.printStackTrace ();
	    }
	}
    }

    /** Reconfigure the proxy during runtime.
     * @param config the new config.
     */
    public static void reConfigure (Config config) {
	if (instance != null) {
	    instance.setup (config);
	    saveConfig ();
	}
    }
    
    /** Configure the error log.
     */
    private void setupErrorLog () {
	String elog = config.getProperty (getClass ().getName (), "errorlog", "logs/error_log");
	try {
	    //closeLog (errorlog);  // dont close, let it be handled by the gc.
	    errorlog.flush ();
	    if (!elog.equals ("")) {
		File f = new File (elog);
		File p = new File (f.getParent ());
		if (!p.exists ())
		    p.mkdirs ();
		errorlog = new PrintWriter (new FileWriter (elog, true),true);       
	    }
	    else 
		errorlog = new PrintWriter (System.err, true);
	} catch (IOException e) {
	    logError (FATAL, "Could not create error log on '" + elog +  "' exiting");
	    System.exit (-1);
	}
	config.setProperty (getClass ().getName (), "errorlog", elog);
    }
    
    /** Configure the access log.
     */
    private void setupAccessLog () {
	String alog = config.getProperty (getClass ().getName (), "accesslog", "logs/access_log");
	try {
	    //closeLog (accesslog);
	    accesslog.flush ();
	    if (!alog.equals ("")) {
		File f = new File (alog);
		File p = new File (f.getParent ());
		if (!p.exists ())
		    p.mkdirs ();		
		accesslog = new PrintWriter (new FileWriter (alog, true),true);       
	    }
	    else 
		accesslog = new PrintWriter (System.out, true);
	} catch (IOException e) {
	    logError (FATAL, "Could not create access log on '" + alog +  "' exiting");
	    System.exit (-1);
	}
	config.setProperty (getClass ().getName (), "accesslog", alog);
    }

    /** Configure the proxy RabbIT is using (if any). 
     */
    private void setupProxyConnection () {
	String pname = config.getProperty (getClass ().getName (), "proxyhost", "");
	String pport = config.getProperty (getClass ().getName (), "proxyport", "");
	if (!pname.equals ("") && !pport.equals ("")) {
	    try {
		proxy = InetAddress.getByName (pname);
	    } catch (UnknownHostException e) {
		logError (FATAL, "Unknown proxyhost: '" + pname + "' exiting");
		System.exit (-1);
	    }
	    try {
		proxyport = Integer.parseInt (pport.trim ());	
	    } catch (NumberFormatException e) {
		logError (FATAL, "Strange proxyport: '" + pport + "' exiting");
		System.exit (-1);
	    }
	}
	config.setProperty (getClass ().getName (), "proxyhost", pname);
	config.setProperty (getClass ().getName (), "proxyport", pport);
    }
    
    /** Configure the SSL support RabbIT should have.
     */
    private void setupSSLSupport () {
	String ssl = config.getProperty (getClass ().getName (), "allowSSL", "no").trim ();
	if (ssl.equals ("no")) {
	    proxySSL = false;
	} else if (ssl.equals ("yes")) {
	    proxySSL = true;
	    sslports = null;
	} else {
	    proxySSL = true;	    
	    // ok, try to get the portnumbers.
	    sslports = new Vector ();
	    StringTokenizer st = new StringTokenizer (ssl, ",");
	    while (st.hasMoreTokens ()) {
		String s = null;
		try {
		    sslports.addElement (new Integer (s = st.nextToken ()));
		} catch (NumberFormatException e) {
		    logError (WARN, "bad number: '" + s + "' for ssl port, ignoring.");
		}
	    }
	}
    }

    /** Configure the maximum number of simultanious connections we handle 
     */
    private void setupMaxConnections () {
	String mc = config.getProperty (getClass ().getName (), "maxconnections", "500").trim ();
	try {
	    maxconnections=Integer.parseInt (mc);
	    tp.setLimit (maxconnections);
	} catch (NumberFormatException e) {
	    logError (WARN, "bad number for maxconnections: '" + mc + "', using old value: " + maxconnections);
	}
    }

    /** set things up according to the config
     * @param config the new config.
     */
    private void setup (Config config) {
	this.config = config;
	/* First we set things up as they should be...
	 * These first step are crucial, we can not reliably recover from
	 * errors here so exit on failure. (Some people may disaggre, but its my call /robo).
	 */

	String loglvl = config.getProperty (getClass ().getName (), "loglevel", "MSG");
	loglevel = getErrorLevel (loglvl);
	
	setupErrorLog ();
	setupAccessLog ();
	openSocket ();
	setupProxyConnection ();

	try {
	    cache.setup (config.getProperties (cache.getClass ().getName ()));
	} catch (IllegalConfigurationException e) {
	    Proxy.logError (ERROR, e.getMessage ());
	}

	try {
	    conhandler.setup (config.getProperties (conhandler.getClass ().getName ()));
	} catch (IllegalConfigurationException e) {
	    Proxy.logError (ERROR, e.getMessage ());
	}
	
	/* From now on we should be able to run without fatal errors... */
	loadClasses ();
	setupMaxConnections ();	
	setupSSLSupport ();
	serverIdentity = config.getProperty (getClass ().getName (), "serverIdentity", VERSION);	
	logError (MSG, "Configuration loaded, ready for action.");
    }
    
    /** save the config back to file
     */
    public static synchronized void saveConfig (){
	try {
	    FileOutputStream fos = new FileOutputStream (configfile);
	    config.save (fos, "This file was automatically generated at " + sdf.format (new Date ()));
	} catch (IOException e) {
	    logError (ERROR, "Coulnd not write the configuration file: '" + configfile + "' to disk");
	}
    }

    /** Get the actual error level from the given String.
     * @param errorlevel the String to translate.
     * @return the errorlevel suitable for the given String.
     */
    public static int getErrorLevel (String errorlevel) {
	if (errorlevel.equals ("DEBUG"))
	    return DEBUG;
	if (errorlevel.equals ("ALL"))
	    return ALL;
	if (errorlevel.equals ("INFO"))
	    return INFO;
	if (errorlevel.equals ("WARN"))
	    return WARN;
	if (errorlevel.equals ("MSG"))
	    return MSG;
	if (errorlevel.equals ("ERROR"))
	    return ERROR;
	if (errorlevel.equals ("FATAL"))
	    return FATAL;
	return FATAL;
    }

    /** Get the String description of the given error level
     * @param errorlevel the int to translate to a String.
     * @return the String describing the errorlevel.
     */
    public static String getErrorLevelString (int errorlevel) {
	if (errorlevel <= DEBUG)
	    return "DEBUG";
	if (errorlevel <= ALL)
	    return "ALL";
	if (errorlevel <= INFO)
	    return "INFO";
	if (errorlevel <= WARN)
	    return "WARN";
	if (errorlevel <= MSG)
	    return "MSG";
	if (errorlevel <= ERROR)
	    return "ERROR";
	if (errorlevel <= FATAL)
	    return "FATAL";
	return "UNKNOWN";
    }
    
    /** log errors and other important stuff. 
     * @param error the thing that happend.
     */
    public static synchronized void logError (String error) {
	logError (ERROR, error);
    }

    /** log errors and other important stuff. 
     * @param type the type of the error. 
     * @param error the thing that happend.
     */
    public static synchronized void logError (int type, String error) {
	if (type < loglevel)
	    return;
	String stype = getErrorLevelString (type);
	Date d = new Date ();
	d.setTime (d.getTime () - offset);
	errorlog.println ("[" + sdf.format (d) + "][" + stype + "][" + error + "]");
    }

    /** Get the cache the proxy are using.
     * @return the NCache the proxy currently are using.
     */
    public static NCache getCache () {
	return cache;
    }

    /** Get the Log-handle
     * @return the Counter we use.
     */ 
    public static Counter getCounter () {
	return log;
    }

    /** Get the port this proxy is using.
     * @return the port number the proxy is listening on.
     */ 
    public static int getPort () {
	return port;
    }
    
    /** Get the local host.
     * @return the InetAddress of the host the proxy is running on.
     */
    public static InetAddress getHost () {
	return localhost;
    }

    /** are we connected to another proxy? 
     * @return true if the proxy is connected to another proxy.
     */
    public static boolean isProxyConnected () {
	return proxy != null;
    }

    /** Get the InetAddress to connect to.
     * @return the InetAddress to give the request
     */
    public static InetAddress getInetAddress (URL url) throws UnknownHostException {
	if (isProxyConnected ())
	    return proxy;
	return InetAddress.getByName(url.getHost());	
    }
    
    /** Get the port to connect to.
     * @param port the port we want to connect to.
     * @return the port to connect to.
     */
    public static int getConnectPort (int port) {
	if (isProxyConnected ())    // are we talking through another proxy?
	    return proxyport;
	return port;
    }
    
    /** Get the authenticationstring to use for proxy.
     * @return an authentication string.
     */ 
    public static String getProxyAuthString () {
	return config.getProperty (Proxy.class.getName (), "proxyauth");
    }

    /** returns the current connections.
     * @return a Vector with the current connections. 
     */
    public static Vector getCurrentConections () {
	return convec;
    }

    /** Remove a connection (it has received full treatment)
     *  @param con the connection that is closing down.
     */
    public static void removeConnection (Connection con) {
	convec.removeElement (con);
	tp.returnThread (con);
    }

    /** return the time when the proxy was started.
     * @return a Date with the time the proxy was started.
     */
    public static Date getStartDate () {
	return started;
    }

    public static String getServerIdentity () {
	return serverIdentity;
    }

    /** load a set of handlers.
     * @param section the section in the config file.
     * @return a Hashtable with mimetypes as keys and Handlers as values.
     */
    protected Hashtable loadHandlers (String section) {
	Properties handlers = config.getProperties (section);
	Enumeration e = handlers.keys ();
	String handler = "", classname = "";
	Hashtable hhandlers = new Hashtable ();
	Hashtable chandlers = new Hashtable ();
	while (e.hasMoreElements ()) {
	    try {
		handler = (String)e.nextElement ();
		classname = handlers.getProperty (handler).trim ();
		Class cls = Class.forName (classname);
		// only call setup once per class...
		if (chandlers.get (cls) == null) {
		    Class sparams[] = new Class[1];
		    sparams[0] = Properties.class;
		    Object sarguments[] = new Object [1];
		    sarguments[0] = config.getProperties (classname);
		    callMethod (cls, "setup", sparams, sarguments);		    
		} 
		chandlers.put (cls, cls);		
		hhandlers.put (handler, cls);
	    } catch (ClassNotFoundException ex) {
		logError (ERROR, "Could not load class: '" + classname + "' for handler '" + handler + "'");
	    }
	}
	return hhandlers;
    }

    /** Make sure all filters and handlers are available
     */
    protected void loadClasses () {
	Proxy.handlers = loadHandlers ("Handlers");
	Proxy.cachehandlers = loadHandlers ("CacheHandlers");
	
	String filters = config.getProperty ("Filters", "accessfilters","");
	Class[] params = new Class[1];
	params[0] = java.net.Socket.class;
	accessfilters = new Vector ();
	loadMethods (filters, accessfilters, "doIPFiltering", params);
	
	filters = config.getProperty ("Filters", "httpinfilters","");
	params = new Class[3];
	params[0] = java.net.Socket.class;
	params[1] = rabbit.http.HTTPHeader.class;
	params[2] = rabbit.proxy.Connection.class;
	httpinfilters = new Vector ();
	loadMethods (filters, httpinfilters, "doHTTPInFiltering", params);
	
	filters = config.getProperty ("Filters", "httpoutfilters","");
	httpoutfilters = new Vector ();
	loadMethods (filters, httpoutfilters, "doHTTPOutFiltering", params);	
    }

    /** loads the methods for a filtergroup.
     *  loads the classes needed and calls setup on them.  
     */
    private void loadMethods (String filters, Vector vector, String methodname, Class[] params) {	
	StringTokenizer st = new StringTokenizer (filters, ",");
	String classname = "";
	while (st.hasMoreElements ()) {
	    try {
		classname = st.nextToken ().trim ();
		Class cls = Class.forName (classname);		
		Method m = cls.getDeclaredMethod (methodname, params);
		vector.addElement (m);

		Class sparams[] = new Class[1];
		sparams[0] = Properties.class;
		Object sarguments[] = new Object [1];
		sarguments[0] = config.getProperties (classname);
		callMethod (cls, "setup", sparams, sarguments);

	    } catch (ClassNotFoundException ex) {
		logError (ERROR, "Could not load class: '" + classname + "' " + ex);
	    } catch (NoSuchMethodException ex) {
		logError (ERROR, "Could not setup class: '" + classname + "' " + ex);
	    }
	}
    }

    /** Call a method. Mostly used to call the static "setup" methods.
     */
    private void callMethod (Class cls, String name, Class[] params, Object[] arguments) {
	try {
	    Method setup = cls.getDeclaredMethod (name, params);	    
	    setup.invoke (cls, arguments);	      
	} catch (NoSuchMethodException e) {
	    logError (WARN, "Could not call method: '" + name + "' for class: '" + cls.getName() + "' " + e);
	} catch (InvocationTargetException e) {
	    logError (WARN, "Could not call method: '" + name + "' for class: '" + cls.getName() + "' " + e);
	} catch (IllegalAccessException e) {
	    logError (WARN, "Could not call method: '" + name + "' for class: '" + cls.getName() + "' " + e);	    
	}
    }
    
    /** loads the classes for a filtergroup.
     */
    private void loadFilters (String filters, Vector vector, String filter) {
	StringTokenizer st = new StringTokenizer (filters, ",");
	String classname = "";
	while (st.hasMoreElements ()) {
	    try {
		classname = st.nextToken ().trim ();
		Class cls = Class.forName (classname);		
		vector.addElement (cls);
	    } catch (ClassNotFoundException e) {
		logError (ERROR, "Could not load class: '" + classname + "' for '" + filter + "'");		
	    }	    
	}
    }
    
    /** log a connection handled.
     * @param con the Connection that handled a request
     */
    public static synchronized void logConnection (Connection con) {
	log.inc ("Total pages served");

	Date d = new Date ();
	d.setTime (d.getTime () - offset);
	
	StringBuffer sb = new StringBuffer (con.getSocket ().getInetAddress ().getHostAddress().toString());
	sb.append (" - ");
	sb.append ((con.getUserName () != null ? con.getUserName () : "-"));
	sb.append (" ");
	sb.append (sdf.format (d));
	sb.append (" \"");
	sb.append (con.getRequestLine ());
	sb.append ("\" ");
	sb.append (con.getStatusCode ());
	sb.append (" ");
	sb.append (con.getContentLength ());
	sb.append (" ");
	sb.append ((con.getExtraInfo () != null ? con.getExtraInfo () : ""));
	accesslog.println (sb.toString ());
    }

    /** shutdown
     */
    public static void kill () {
	closeSocket ();
	accesslog.flush ();
	accesslog.close ();
	logError (MSG, "Kill issued, shuting down");
	errorlog.flush ();
	errorlog.close ();
	cache.flush ();
	System.exit (0);
    }
}


