package rabbit.util;

// $Id: Access.java,v 1.1.1.1 1999/12/18 18:49:32 robo Exp $

import java.net.*;

/** A class to handle access to ip ranges.
 */
public class Access {
    /** No access at all */
    public static int NONE = 0;
    /** allowed to access metapages */
    public static int ADMIN = 1;
    /** allowed to access standard http (ordinary webpages) */
    public static int HTTP = 2;
    /** allowed to run secure transactions */
    public static int HTTPS = 4;
    /** allowed to ftp */
    public static int FTP = 8;
    /** allowed to run gopher */
    public static int GOPHER = 16;
    /** allowed to run wais */
    public static int WAIS = 32;
  
    private long lowip;          // lowest ip in range
    private long highip;         // highest ip in range
    private int type = NONE;

    /** Create a new Access with given ip-range and given type
     * @param type one of http, admin, ftp, gopher, https, wais (caseinsensitive)
     * @param lowip the lowest ip in the range
     * @param highip the highest ip in the range
     */
    public Access (String type, InetAddress lowip, InetAddress highip) {
	type = type.toLowerCase ();
	int itype = NONE;
	if (type.equals ("http"))
	    itype = HTTP;
	if (type.equals ("admin"))
	    itype = ADMIN;
	if (type.equals ("ftp"))
	    itype = FTP;
	if (type.equals ("gopher"))
	    itype = GOPHER;
	if (type.equals ("https"))
	    itype = HTTPS;
	if (type.equals ("wais"))
	    itype = WAIS;
	if (lowip == null || highip == null)
	    return;
	setup (itype, lowip, highip);
    }

    /** Create a new Access with given ip-range and given type
     * @param type one of the constants (HTTP, FTP, ADMIN etc) or a OR (|) of them. 
     * @param lowip the lowest ip in the range
     * @param highip the highest ip in the range
     */
    public Access (int type, InetAddress lowipa, InetAddress highipa) {
	if (lowipa == null || highipa == null)
	    return;
	setup (type, lowipa, highipa);
    }

    /** transform to suitable variabletypes
     */
    private void setup (int type, InetAddress lowipa, InetAddress highipa) {
	this.type = type;
	lowip = getLongFromIP (lowipa);
	highip = getLongFromIP (highipa);

	if (lowip > highip) {
	    long t = lowip;
	    lowip = highip;
	    highip = lowip;
	} 
    }
  
    /** make an long from the ip so we can do simple test later on
     */
    private long getLongFromIP (InetAddress ia) {
	byte[] address = ia.getAddress ();
	long ret = 0;
	for (int i = 0; i < address.length; i++) {
	    ret <<= 8;
	    long a = address[i] & 0xff; // byte is signed, signextension is evil..
	    ret |= a;
	}
	return ret;
    }

    /** check if a given ip is in this accessrange
     * @param ia the ip we are testing.
     * @return true if ia is in the range (inclusive), false otherwise 
     */ 
    public boolean inrange (InetAddress ia) {
	long check = getLongFromIP (ia);
	return (check >= lowip && check <= highip);
    }

    /** return the accesstype we have for this range 
     */
    public int getType () {
	return type;
    }
    /** sets the accesstype we have for this range 
     * @param newType the new access for this range.
     */
    public void setType (int newType) {
	type = newType;
    }
    
    /** get the string representation of this access.
     */
    public String toString () {
	String acc = "";
	if ((type & NONE) > 0) 
	    acc += "NONE,";
	if ((type & ADMIN) > 0)
	    acc += "ADMIN,";
	if ((type & HTTP) > 0)
	    acc += "HTTP,";	
	if ((type & HTTPS) > 0)
	    acc += "HTTPS,";
	if ((type & FTP) > 0)
	    acc += "FTP,";
	if ((type & GOPHER) > 0)
	    acc += "GOPHER,";
	if ((type & WAIS) > 0)
	    acc += "WAIS,";
	if (!acc.equals (""))
	    acc = acc.substring (0, acc.length() - 1);
	
	String low = getIP (lowip);
	String high = getIP (highip);
	
	return acc + "\t" + low + "\t" + high;
    }
    
    private String getIP (long ip) {
	return  "" + (ip>>24) + "." +
	    ((ip>>16)&0xff) + "." +
	    ((ip>>8)&0xff) + "." +
	    (ip&255);
    }
}
