package rabbit.util;

// $Id: ThreadPool.java,v 1.1 2000/12/27 19:06:58 robo Exp $ 

import java.util.*;

/** This is a thread pool. It uses RestartableThread.
 */
public class ThreadPool {
    
    private Vector free_threads;
    private Vector used_threads;
    private RestartableThreadFactory tf;
    private int limit = Integer.MAX_VALUE;

    /** Create a ThreadPool.
     * @param tf the RestartableThreadFactory that is used as more Threads are needed.
     * @param limit dont create more threads than this.
     */
    public ThreadPool (RestartableThreadFactory tf, int limit) {
	this.tf = tf;
	this.limit = limit;
	free_threads = new Vector ();
	used_threads = new Vector ();	
    }
    
    /** Set the limit.
     * @param limit the new limit of concurrent threads.
     */
    public void setLimit (int limit) {
	this.limit = limit;
    }

    /** Get the current limit of concurrent threads.
     * @return the limit of threads.
     */
    public int getLimit () {
	return limit;
    }

    /** Get a Thread. This will wait for a thread 
     *  to be returned if the limit is currently reached.
     */
    public synchronized RestartableThread getThread () {
	while (free_threads.size () == 0 && 
	       used_threads.size () == limit) {
	    try {
		wait ();   // wait for a thread to become free.
	    } catch (InterruptedException e) {
		// ignore.
	    }
	}
	
	if (free_threads.size () == 0) {
	    // we know that used_threads.size () < limit now...
	    RestartableThread t = tf.createThread ();
	    used_threads.addElement (t);
	    return t;
	} else {
	    RestartableThread t = (RestartableThread)free_threads.lastElement ();
	    free_threads.removeElementAt (free_threads.size () - 1);
	    used_threads.addElement (t);
	    return t;
	}
    }

    /** Return a thread to the pool.
     *  If the thread was never part of this pool it will be ignored.
     * @param t the thread to return.
     */
    public synchronized void returnThread (RestartableThread t) {
	int i = used_threads.indexOf (t);
	if (i == -1) {
	    System.err.println ("Unknown thread: " + t);
	    return;
	}
	boolean shouldnotify = ((free_threads.size () == 0) && 
				(used_threads.size () == limit));
	used_threads.removeElementAt (i);
	free_threads.addElement (t);
	if (shouldnotify)
	    notify ();
    }

    /** Set the factory to use for creating threads.
     * @param tf the new factory.
     */
    public void setFactory (RestartableThreadFactory tf) {
	this.tf = tf;
    }
    
    /** Get the current factory.
     * @return the current factory.
     */
    public RestartableThreadFactory getFactory () {
	return tf;
    }
}
